#!/usr/bin/env python
#
# (C)2011 Red Hat, Inc., Lukas Czerner <lczerner@redhat.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

# Unittests for the system storage manager lvm backend

import unittest
from ssmlib import main
from ssmlib import problem
from ssmlib.backends import lvm
from tests.unittests.common import *


class LvmFunctionCheck(MockSystemDataSource):

    def setUp(self):
        super(LvmFunctionCheck, self).setUp()
        self._addDevice('/dev/sda', 11489037516)
        self._addDevice('/dev/sdb', 234566451)
        self._addDevice('/dev/sdc', 2684354560)
        self._addDevice('/dev/sdc1', 894784853, 1)
        self._addDevice('/dev/sdc2', 29826161, 2)
        self._addDevice('/dev/sdc3', 1042177280, 3)
        self._addDevice('/dev/sdd', 11673)
        self._addDevice('/dev/sde', 1042177280)
        main.SSM_DEFAULT_BACKEND = 'lvm'

    def mock_run(self, cmd, *args, **kwargs):

        # Convert all parts of cmd into string
        for i, item in enumerate(cmd):
            if type(item) is not str:
                cmd[i] = str(item)

        self.run_data.append(" ".join(cmd))
        output = ""
        if cmd[1] == 'pvs':
            for (dev, data) in self.dev_data.items():
                if 'pool_name' in data:
                    output += "{0}|{1}|{2}|{3}\n".format(dev, data['pool_name'],
                            data['dev_free'], data['dev_used'])
        elif cmd[1] == 'vgs':
            for (pool, data) in self.pool_data.items():
                output += "{0}|{1}|{2}|{3}|{4}\n".format(pool, data['dev_count'],
                        data['pool_size'], data['pool_free'], data['vol_count'])
        elif cmd[1] == 'lvs':
            for (vol, data) in self.vol_data.items():
                output += "{0}|{1}|{2}|{3}|{4}|{5}|{6}|{7}\n".format(data['pool_name'],
                        data['vol_size'], data['stripes'], data['stripesize'],
                        data['type'], data['dev_name'].split("/")[-1],
                        data['origin'], data['attr'])
        if 'return_stdout' in kwargs and not kwargs['return_stdout']:
            output = None
        return (0, output, None)

    def test_lvm_create(self):
        default_pool = lvm.SSM_LVM_DEFAULT_POOL

        # Create volume using single device from non existent default pool
        self._checkCmd("ssm create", ['/dev/sda'],
            "lvm lvcreate {0} -l 100%PVS -n lvol001 /dev/sda".format(default_pool))
        self._cmdEq("lvm vgcreate {0} /dev/sda".format(default_pool), -2)

        # Specify backnend
        self._checkCmd("ssm -b lvm create", ['/dev/sda'],
            "lvm lvcreate {0} -l 100%PVS -n lvol001 /dev/sda".format(default_pool))
        self._cmdEq("lvm vgcreate {0} /dev/sda".format(default_pool), -2)

        main.SSM_DEFAULT_BACKEND = "btrfs"
        self._checkCmd("ssm --backend lvm create", ['/dev/sda'],
            "lvm lvcreate {0} -l 100%PVS -n lvol001 /dev/sda".format(default_pool))
        self._cmdEq("lvm vgcreate {0} /dev/sda".format(default_pool), -2)
        main.SSM_DEFAULT_BACKEND = "lvm"

        self._checkCmd("ssm create", ['--name myvolume', '--fstype ext4', '/dev/sda'])
        self._cmdEq("mkfs.ext4 /dev/{0}/myvolume".format(default_pool))
        self._cmdEq("lvm lvcreate {0} -l 100%PVS -n myvolume /dev/sda".format(default_pool), -2)
        self._cmdEq("lvm vgcreate {0} /dev/sda".format(default_pool), -3)

        self._checkCmd("ssm -f create", ['--fstype ext4', '/dev/sda'])
        self._cmdEq("mkfs.ext4 -F /dev/{0}/lvol001".format(default_pool))
        self._cmdEq("lvm lvcreate {0} -l 100%PVS -n lvol001 /dev/sda".format(default_pool), -2)
        self._cmdEq("lvm vgcreate -f {0} /dev/sda".format(default_pool), -3)

        self._checkCmd("ssm -v create", ['--name myvolume', '--fstype xfs', '/dev/sda'])
        self._cmdEq("mkfs.xfs /dev/{0}/myvolume".format(default_pool))
        self._cmdEq("lvm lvcreate -v {0} -l 100%PVS -n myvolume /dev/sda".format(default_pool), -2)
        self._cmdEq("lvm vgcreate -v {0} /dev/sda".format(default_pool), -3)

        self._checkCmd("ssm -v -f create", ['--name myvolume', '--fstype xfs', '/dev/sda'])
        self._cmdEq("mkfs.xfs -f /dev/{0}/myvolume".format(default_pool))
        self._cmdEq("lvm lvcreate -v {0} -l 100%PVS -n myvolume /dev/sda".format(default_pool), -2)
        self._cmdEq("lvm vgcreate -v -f {0} /dev/sda".format(default_pool), -3)

        self._checkCmd("ssm create", ['-s 2.6T', '/dev/sda'],
            "lvm lvcreate {0} -L 2791728742.4K -n lvol001 /dev/sda".format(default_pool))
        self._cmdEq("lvm vgcreate {0} /dev/sda".format(default_pool), -2)

        self._checkCmd("ssm create", ['-r 0', '-s 2.6T', '-I 16', '/dev/sda'],
            "lvm lvcreate {0} -L 2791728742.4K -n lvol001 -I 16 -i 1 /dev/sda".format(default_pool))
        self._cmdEq("lvm vgcreate {0} /dev/sda".format(default_pool), -2)

        self._checkCmd("ssm create", ['-r 0', '-s 2.6T', '-I 16', '/dev/sda'],
            "lvm lvcreate {0} -L 2791728742.4K -n lvol001 -I 16 -i 1 /dev/sda".format(default_pool))
        self._cmdEq("lvm vgcreate {0} /dev/sda".format(default_pool), -2)

        # Number of stripes must not exceed number of devices
        self.assertRaises(problem.GeneralError, main.main, "ssm create -r 0 -s 2.6T -I 16 -i 4 /dev/sda")

        # Create volume using single device from non existent my_pool
        self._checkCmd("ssm create", ['--pool my_pool', '/dev/sda'],
            "lvm lvcreate my_pool -l 100%PVS -n lvol001 /dev/sda")
        self._cmdEq("lvm vgcreate my_pool /dev/sda", -2)

        self._checkCmd("ssm create", ['-r 0', '-p my_pool', '-s 2.6T', '-I 16',
            '-i 2', '/dev/sda /dev/sdb'],
            "lvm lvcreate my_pool -L 2791728742.4K -n lvol001 -I 16 -i 2 /dev/sda /dev/sdb")
        self._cmdEq("lvm vgcreate my_pool /dev/sda /dev/sdb", -2)

        # Create volume using multiple devices
        self._checkCmd("ssm create", ['/dev/sda /dev/sdc1'],
            "lvm lvcreate {0} -l 100%PVS -n lvol001 /dev/sda /dev/sdc1".format(default_pool))
        self._cmdEq("lvm vgcreate {0} /dev/sda /dev/sdc1".format(default_pool), -2)


        # Create volume using single devices from existing pool
        self._addPool(default_pool, ['/dev/sdb', '/dev/sdd'])

        self._checkCmd("ssm create", ['-r 0', '-s 2.6T', '-I 16',
            '-n myvolume', '/dev/sda'],
            "lvm lvcreate {0} -L 2791728742.4K -n myvolume -I 16 -i 1 /dev/sda". format(default_pool))
        self._cmdEq("lvm vgextend {0} /dev/sda".format(default_pool), -2)

        self._addPool("my_pool", ['/dev/sdc2', '/dev/sdc3'])
        self._checkCmd("ssm create", ['-r 0', '-p my_pool', '-s 2.6T', '-I 16',
            '-n myvolume', '/dev/sda'],
            "lvm lvcreate my_pool -L 2791728742.4K -n myvolume -I 16 -i 1 /dev/sda")
        self._cmdEq("lvm vgextend my_pool /dev/sda", -2)

        # Create volume using multiple devices which one of the is in already
        # in the pool
        self._checkCmd("ssm create", ['-n myvolume', '/dev/sda /dev/sdb'],
            "lvm lvcreate {0} -l 100%PVS -n myvolume /dev/sda /dev/sdb". format(default_pool))
        self._cmdEq("lvm vgextend {0} /dev/sda".format(default_pool), -2)

        self._addPool("my_pool", ['/dev/sdc2', '/dev/sdc3'])
        self._checkCmd("ssm create", ['-p my_pool', '-n myvolume', '/dev/sdc2 /dev/sda'],
            "lvm lvcreate my_pool -l 100%PVS -n myvolume /dev/sdc2 /dev/sda")
        self._cmdEq("lvm vgextend my_pool /dev/sda", -2)

        self._checkCmd("ssm create", ['-n myvolume', '/dev/sda /dev/sdb /dev/sde'],
            "lvm lvcreate {0} -l 100%PVS -n myvolume /dev/sda /dev/sdb /dev/sde". format(default_pool))
        self._cmdEq("lvm vgextend {0} /dev/sda /dev/sde".format(default_pool), -2)

    def test_lvm_remove(self):
        # Generate some storage data
        self._addPool('default_pool', ['/dev/sda', '/dev/sdb'])
        self._addPool('my_pool', ['/dev/sdc2', '/dev/sdc3', '/dev/sdc1'])
        self._addVol('vol001', 117283225, 1, 'default_pool', ['/dev/sda'])
        self._addVol('vol002', 237284225, 1, 'default_pool', ['/dev/sda'])
        self._addVol('vol003', 1024, 1, 'default_pool', ['/dev/sdd'])
        self._addVol('vol004', 209715200, 2, 'default_pool', ['/dev/sda',
                     '/dev/sdb'])

        # remove volume
        main.main("ssm remove /dev/default_pool/vol002")
        self._cmdEq("lvm lvremove /dev/default_pool/vol002")
        # remove multiple volumes
        main.main("ssm remove /dev/default_pool/vol002 /dev/default_pool/vol003")
        self.assertEqual(self.run_data[-2], "lvm lvremove /dev/default_pool/vol002")
        self._cmdEq("lvm lvremove /dev/default_pool/vol003")
        # remove pool
        main.main("ssm remove my_pool")
        self._cmdEq("lvm vgremove my_pool")
        # remove multiple pools
        main.main("ssm remove my_pool default_pool")
        self.assertEqual(self.run_data[-2], "lvm vgremove my_pool")
        self._cmdEq("lvm vgremove default_pool")
        # remove device
        main.main("ssm remove /dev/sdc1")
        self._cmdEq("lvm vgreduce my_pool /dev/sdc1")
        # remove multiple devices
        main.main("ssm remove /dev/sdc1 /dev/sdb")
        self.assertEqual(self.run_data[-2], "lvm vgreduce my_pool /dev/sdc1")
        self._cmdEq("lvm vgreduce default_pool /dev/sdb")
        # remove combination
        main.main("ssm remove /dev/sdb my_pool /dev/default_pool/vol001")
        self.assertEqual(self.run_data[-3], "lvm vgreduce default_pool /dev/sdb")
        self.assertEqual(self.run_data[-2], "lvm vgremove my_pool")
        self._cmdEq("lvm lvremove /dev/default_pool/vol001")
        # remove all
        main.main("ssm remove --all")
        self.assertEqual(self.run_data[-2], "lvm vgremove default_pool")
        self._cmdEq("lvm vgremove my_pool")
        # remove force
        main.main("ssm -f remove /dev/default_pool/vol002")
        self._cmdEq("lvm lvremove -f /dev/default_pool/vol002")
        # remove verbose
        main.main("ssm -v remove /dev/default_pool/vol002")
        self._cmdEq("lvm lvremove -v /dev/default_pool/vol002")
        # remove verbose + force
        main.main("ssm -v -f remove /dev/default_pool/vol002")
        self._cmdEq("lvm lvremove -v -f /dev/default_pool/vol002")

    def test_lvm_snapshot(self):
        # Generate some storage data
        self._addPool('default_pool', ['/dev/sda', '/dev/sdb'])
        self._addPool('my_pool', ['/dev/sdc2', '/dev/sdc3', '/dev/sdc1'])
        self._addVol('vol001', 117283225, 1, 'default_pool', ['/dev/sda'])
        self._addVol('vol002', 237284225, 1, 'default_pool', ['/dev/sda'],
                    '/mnt/mount1')
        self._addVol('vol003', 1024, 1, 'default_pool', ['/dev/sdd'])
        self._addVol('vol004', 209715200, 2, 'default_pool', ['/dev/sda',
                     '/dev/sdb'], '/mnt/mount')

        # Create snapshot
        self._checkCmd("ssm snapshot --name new_snap", ['/dev/default_pool/vol001'],
            "lvm lvcreate --size 23456645.0K --snapshot --name new_snap /dev/default_pool/vol001")

        main.SSM_DEFAULT_BACKEND = "btrfs"
        self._checkCmd("ssm snapshot --name new_snap", ['/dev/default_pool/vol001'],
            "lvm lvcreate --size 23456645.0K --snapshot --name new_snap /dev/default_pool/vol001")
        main.SSM_DEFAULT_BACKEND = "lvm"

        # Create snapshot verbose
        self._checkCmd("ssm -v snapshot --name new_snap", ['/dev/default_pool/vol001'],
            "lvm lvcreate -v --size 23456645.0K --snapshot --name new_snap /dev/default_pool/vol001")
        # Create snapshot force
        self._checkCmd("ssm -f snapshot --name new_snap", ['/dev/default_pool/vol001'],
            "lvm lvcreate -f --size 23456645.0K --snapshot --name new_snap /dev/default_pool/vol001")
        # Create snapshot force verbose
        self._checkCmd("ssm -f -v snapshot --name new_snap", ['/dev/default_pool/vol001'],
            "lvm lvcreate -v -f --size 23456645.0K --snapshot --name new_snap /dev/default_pool/vol001")

        # Create snapshot with size and name specified
        self._checkCmd("ssm snapshot", ['--size 1G', '--name new_snap',
                                        '/dev/default_pool/vol001'],
            "lvm lvcreate --size 1048576.0K --snapshot --name new_snap /dev/default_pool/vol001")

    def test_lvm_migrate(self):
        # Generate some storage data
        self._addDevice('/dev/sdd1', 11489037516)
        self._addDevice('/dev/sdd2', 11489037516)
        self._addDevice('/dev/sde', 11489037516)
        self._addPool('my_pool', ['/dev/sdc1', '/dev/sdc2'])
        self._addVol('vol002', 237284225, 1, 'my_pool', ['/dev/sdc1'])
        self._addPool('my_pool2', ['/dev/sdd1', '/dev/sdd2'])
        self._addVol('vol002', 237284225, 1, 'my_pool2', ['/dev/sdd1'])

        self._checkCmd("ssm migrate /dev/sdc1 /dev/sdc2", [],
            "lvm pvmove --atomic /dev/sdc1 /dev/sdc2")
        self._checkCmd("ssm migrate /dev/sdc1 /dev/sde", [],
            "lvm pvmove --atomic /dev/sdc1 /dev/sde")
        with self.assertRaises(problem.UserInterrupted):
            main.main("ssm migrate /dev/sde /dev/sdc2")

        # test moving a pv from one pool to another
        with self.assertRaises(problem.UserInterrupted):
            main.main("ssm migrate /dev/sdc1 /dev/sdd2")
        self._checkCmd("ssm -f migrate /dev/sdc1 /dev/sdd2", [],
            "lvm pvmove --atomic /dev/sdc1 /dev/sdd2")
        self.assertTrue("lvm vgreduce -f my_pool2 /dev/sdd2" in self.run_data)
        self.assertTrue("lvm vgextend -f my_pool /dev/sdd2" in self.run_data)


    def test_lvm_resize(self):
        # Generate some storage data
        self._addPool('default_pool', ['/dev/sda', '/dev/sdb'])
        self._addPool('my_pool', ['/dev/sdc2', '/dev/sdc3'])
        self._addVol('vol001', 2982616, 1, 'my_pool', ['/dev/sdc2'],
                    '/mnt/test1')
        self._addVol('vol002', 237284225, 1, 'default_pool', ['/dev/sda'])
        self._addVol('vol003', 1024, 1, 'default_pool', ['/dev/sdd'])
        self._addDevice('/dev/sde', 11489037516)

        # Extend Volume
        self._checkCmd("ssm resize", ['--size +4m', '/dev/default_pool/vol003'],
            "lvm lvresize -L 5120.0k /dev/default_pool/vol003")

        # Specify backend
        self._checkCmd("ssm --backend lvm resize", ['--size +4m', '/dev/default_pool/vol003'],
            "lvm lvresize -L 5120.0k /dev/default_pool/vol003")

        main.SSM_DEFAULT_BACKEND = "btrfs"
        self._checkCmd("ssm resize", ['--size +4m', '/dev/default_pool/vol003'],
            "lvm lvresize -L 5120.0k /dev/default_pool/vol003")
        main.SSM_DEFAULT_BACKEND = "lvm"

        # Shrink volume
        self._checkCmd("ssm resize", ['-s-100G', '/dev/default_pool/vol002'],
            "lvm lvresize -L 132426625.0k /dev/default_pool/vol002")

        # Set volume size
        self._checkCmd("ssm resize", ['-s 10M', '/dev/my_pool/vol001'],
            "lvm lvresize -L 10240.0k /dev/my_pool/vol001")

        # Set volume and add devices
        self._checkCmd("ssm resize -s 12T /dev/default_pool/vol003 /dev/sdc1 /dev/sde",
            [], "lvm lvresize -L 12884901888.0k /dev/default_pool/vol003")
        self.assertEqual(self.run_data[-2],
            "lvm vgextend default_pool /dev/sdc1 /dev/sde")

        # Set volume size with sufficient amount of space
        self._checkCmd("ssm resize -s 10G /dev/default_pool/vol003 /dev/sdc1 /dev/sde",
            [], "lvm lvresize -L 10485760.0k /dev/default_pool/vol003")
        self.assertNotEqual(self.run_data[-2],
            "lvm vgextend default_pool /dev/sdc1 /dev/sde")

        # Set volume size without sufficient amount of space
        self._checkCmd("ssm resize -s 10T /dev/default_pool/vol003 /dev/sdc1 /dev/sde",
            [], "lvm lvresize -L 10737418240.0k /dev/default_pool/vol003")
        self.assertNotEqual(self.run_data[-2],
            "lvm vgextend default_pool /dev/sdc1 /dev/sde")

        # Extend volume and add devices
        self._checkCmd("ssm resize -s +11T /dev/default_pool/vol003 /dev/sdc1 /dev/sde",
            [], "lvm lvresize -L 11811161088.0k /dev/default_pool/vol003")
        self.assertEqual(self.run_data[-2],
            "lvm vgextend default_pool /dev/sdc1 /dev/sde")

        # Extend volume with ehough space in pool
        self._checkCmd("ssm resize -s +10G /dev/default_pool/vol003 /dev/sdc1 /dev/sde",
            [], "lvm lvresize -L 10486784.0k /dev/default_pool/vol003")
        self.assertNotEqual(self.run_data[-2],
            "lvm vgextend default_pool /dev/sdc1 /dev/sde")

        # Extend volume without ehough space in pool
        self._checkCmd("ssm resize -s +20T /dev/default_pool/vol003 /dev/sdc1 /dev/sde",
            [], "lvm lvresize -L 21474837504.0k /dev/default_pool/vol003")
        self.assertEqual(self.run_data[-2],
            "lvm vgextend default_pool /dev/sdc1 /dev/sde")

        # Shrink volume with devices provided
        self._checkCmd("ssm resize -s-10G /dev/default_pool/vol002 /dev/sdc1 /dev/sde",
            [], "lvm lvresize -L 226798465.0k /dev/default_pool/vol002")
        self.assertNotEqual(self.run_data[-2],
            "lvm vgextend default_pool /dev/sdc1 /dev/sde")

        # Test that we do not use devices which are already used in different
        # pool
        self.assertRaises(Exception, main.main, "ssm resize -s +1.5T /dev/my_pool/vol001 /dev/sdb /dev/sda")

        # If the device we are able to use can cover the size, then
        # it will be resized successfully
        self._checkCmd("ssm resize -s +1.5T /dev/my_pool/vol001 /dev/sdb /dev/sda /dev/sdc1",
            [], "lvm lvresize -L 1613595352.0k /dev/my_pool/vol001")

        # Test resize on inactive volume
        self._addVol('vol004', 8192, 1, 'default_pool', ['/dev/sdd'], None, False)
        self._checkCmd("ssm resize", ['--size +4m', '/dev/default_pool/vol004'],
            "lvm lvresize -L 12288.0k /dev/default_pool/vol004")
        self.assertRaises(Exception, main.main, "ssm resize -s-2m /dev/default_pool/vol004")
        # We can force it though
        self._checkCmd("ssm -f resize", ['-s-2m', '/dev/default_pool/vol004'],
            "lvm lvresize -f -L 6144.0k /dev/default_pool/vol004")

    def test_lvm_add(self):
        default_pool = lvm.SSM_LVM_DEFAULT_POOL

        # Adding to non existent pool
        # Add device into default pool
        self._checkCmd("ssm add", ['/dev/sda'],
            "lvm vgcreate {0} /dev/sda".format(default_pool))

        # Specify backend
        self._checkCmd("ssm --backend lvm add", ['/dev/sda'],
            "lvm vgcreate {0} /dev/sda".format(default_pool))

        main.SSM_DEFAULT_BACKEND = "btrfs"
        self._checkCmd("ssm --backend lvm add", ['/dev/sda'],
            "lvm vgcreate {0} /dev/sda".format(default_pool))
        main.SSM_DEFAULT_BACKEND = "lvm"

        # Add more devices into default pool
        self._checkCmd("ssm add", ['/dev/sda /dev/sdc1'],
            "lvm vgcreate {0} /dev/sda /dev/sdc1".format(default_pool))
        # Add device into defined pool
        self._checkCmd("ssm add", ['-p my_pool', '/dev/sda'],
            "lvm vgcreate my_pool /dev/sda")
        self._checkCmd("ssm add", ['--pool my_pool', '/dev/sda'],
            "lvm vgcreate my_pool /dev/sda")
        # Add more devices into defined pool
        self._checkCmd("ssm add", ['-p my_pool', '/dev/sda /dev/sdc1'],
            "lvm vgcreate my_pool /dev/sda /dev/sdc1")
        self._checkCmd("ssm add", ['--pool my_pool', '/dev/sda /dev/sdc1'],
            "lvm vgcreate my_pool /dev/sda /dev/sdc1")
        # Add force
        self._checkCmd("ssm -f add", ['--pool my_pool', '/dev/sda'],
            "lvm vgcreate -f my_pool /dev/sda")
        # Add verbose
        self._checkCmd("ssm -v add", ['--pool my_pool', '/dev/sda'],
            "lvm vgcreate -v my_pool /dev/sda")
        # Add force and verbose
        self._checkCmd("ssm -v -f add", ['--pool my_pool', '/dev/sda'],
            "lvm vgcreate -v -f my_pool /dev/sda")

        # Adding to existing default pool
        self._addPool(default_pool, ['/dev/sdb', '/dev/sdd'])
        # Add device into default pool
        self._checkCmd("ssm add", ['/dev/sda'],
            "lvm vgextend {0} /dev/sda".format(default_pool))
        # Add more devices into default pool
        self._checkCmd("ssm add", ['/dev/sda /dev/sdc1'],
            "lvm vgextend {0} /dev/sda /dev/sdc1".format(default_pool))

        # Adding to existing defined pool
        self._addPool("my_pool", ['/dev/sdc2', '/dev/sdc3'])
        # Add device into defined pool
        self._checkCmd("ssm add", ['-p my_pool', '/dev/sda'],
            "lvm vgextend my_pool /dev/sda")
        self._checkCmd("ssm add", ['--pool my_pool', '/dev/sda'],
            "lvm vgextend my_pool /dev/sda")
        # Add more devices into defined pool
        self._checkCmd("ssm add", ['-p my_pool', '/dev/sda /dev/sdc1'],
            "lvm vgextend my_pool /dev/sda /dev/sdc1")
        self._checkCmd("ssm add", ['--pool my_pool', '/dev/sda /dev/sdc1'],
            "lvm vgextend my_pool /dev/sda /dev/sdc1")
        # Add force
        self._checkCmd("ssm -f add", ['--pool my_pool', '/dev/sda'],
            "lvm vgextend -f my_pool /dev/sda")
        # Add verbose
        self._checkCmd("ssm -v add", ['--pool my_pool', '/dev/sda'],
            "lvm vgextend -v my_pool /dev/sda")
        # Add force and verbose
        self._checkCmd("ssm -v -f add", ['--pool my_pool', '/dev/sda'],
            "lvm vgextend -v -f my_pool /dev/sda")

        # Add two devices into existing pool (one of the devices already is in
        # the pool
        self._checkCmd("ssm add", ['--pool my_pool', '/dev/sdc2 /dev/sda'],
            "lvm vgextend my_pool /dev/sda")
        self._checkCmd("ssm add", ['/dev/sda /dev/sdb'],
            "lvm vgextend {0} /dev/sda".format(default_pool))

    def test_lvm_mount(self):
        self._addDir("/mnt/test")
        self._addDir("/mnt/test1")
        self._addDir("/mnt/test2")
        # Generate some storage data
        self._addPool('default_pool', ['/dev/sda', '/dev/sdb'])
        self._addPool('my_pool', ['/dev/sdc2', '/dev/sdc3', '/dev/sdc1'])
        self._addVol('vol001', 117283225, 1, 'default_pool', ['/dev/sda'],
                    '/mnt/test1')
        self._addVol('vol002', 237284225, 1, 'my_pool', ['/dev/sda'])

        # Simple mount test
        main.main("ssm mount /dev/default_pool/vol001 /mnt/test")
        self._cmdEq("mount /dev/default_pool/vol001 /mnt/test")
        main.main("ssm mount -o discard /dev/default_pool/vol001 /mnt/test")
        self._cmdEq("mount -o discard /dev/default_pool/vol001 /mnt/test")
        main.main("ssm mount --options rw,discard,neco=44 /dev/my_pool/vol002 /mnt/test1")
        self._cmdEq("mount -o rw,discard,neco=44 /dev/my_pool/vol002 /mnt/test1")



# (C)2013 Red Hat, Inc., Lukas Czerner <lczerner@redhat.com>
#                        Jan Tulak <jtulak@redhat.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

# multipath backend for ssm

import os
import re
from ssmlib import misc
from ssmlib import problem

from ssmlib.backends import template

__all__ = ["MultipathDevice"]

MP="multipath"

class Multipath(template.Backend):
    def __init__(self, options, data=None):
        self.type = 'multipath'
        self.data = data or {}
        self._dev = {}
        self.options = options
        self.output = None
        self.problem = problem.ProblemSet(options)

        for mp_dev in self.get_mp_devices():
            mpname = self.get_real_device(mp_dev)
            self._dev[mp_dev] = self.get_volume_data(mp_dev)
            for devname in self._dev[mp_dev]['nodes']:
                self._dev[devname] = self.get_device_data(devname, mpname, 0)

    def __str__(self):
        return "mp: %s" % repr(self.data)

    def get_real_device(self, devname):
        """ Get the device for multipath volume name.
            Do we have /dev/mapper/mpathX, or /dev/dm-X?
        """
        if len(devname) > 5 and devname[:5] == "mpath":
            return misc.get_real_device("/dev/mapper/"+devname)
        elif len(devname) > 3 and devname[:3] == "dm-":
            return misc.get_real_device("/dev/"+devname)
        # Or maybe raise an exception?
        return devname


    def get_device_data(self, devname, mpname, devsize):
        data = {}
        data['dev_name'] = devname
        data['hide'] = False
        if mpname:
            data['multipath_volname'] = mpname
            data['pool_name'] = self.get_real_device(mpname)
            data['mount'] = 'MULTIPATH'
        return data

    def get_mp_devices(self):
        """ Find all multipath devices (but not their nodes). """
        devices = []
        command = [MP, '-ll']
        try:
            output = misc.run(command, stderr=False, can_fail=True)[1].split("\n")
            pattern = re.compile(r"^([a-z0-9]+) \([^)]+\)")
        except (problem.CommandFailed, OSError):
            # probably multipath not installed
            output = []

        for line in output:
            match = pattern.match(line)
            if match:
                devices.append(match.group(1))
        return devices


    def get_volume_data(self, volname):
        data = {}
        data['dev_name'] = self.get_real_device(volname)
        data['hide'] = False
        command = [MP, '-ll', volname]
        try:
            output = misc.run(command, stderr=False)[1].split("\n")
        except OSError:
            # probably multipath not installed
            output = []

        if len(output) > 0:
            match = re.search(r"\(([^)]+)\)",output[0])
            data['wwid'] = match.group(1)
            data['dev_size'] = misc.get_device_size(data['dev_name'])
            data['nodes'] = []
            data['total_nodes'] = 0
            for entry in zip(output[2::2],output[3::2]):
                """ Some string operations to remove the tree path symbols
                    from the output. """
                dev = list(filter(None,
                            entry[1][re.search(r"[a-zA-Z0-9]", entry[1])
                                     .start():].split(" ")
                        ))
                data['nodes'].append("/dev/"+self.get_real_device(dev[1]))
                data['total_nodes'] += 1
        return data


class MultipathDevice(Multipath, template.BackendDevice):
    def __init__(self, *args, **kwargs):
        super(MultipathDevice, self).__init__(*args, **kwargs)
        if self.data:
            self.data.update(self._dev)
        else:
            self.data = self._dev



#!/usr/bin/env python
#
# (C)2012 Red Hat, Inc., Lukas Czerner <lczerner@redhat.com>
# (C)2016 Red Hat, Inc., Jan Tulak <jtulak@redhat.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

# Unittests for the system storage manager multipath backend


import unittest
from ssmlib import problem
from ssmlib import main
from ssmlib import misc
from ssmlib.backends import multipath
from tests.unittests.common import *
from os.path import basename
from ssmlib.backends.multipath import MultipathDevice
import sys
try:
    from StringIO import StringIO
except ImportError:
    from io import StringIO

class MultipathFunctionCheck(MockSystemDataSource):


    """
    Create a multipah configuration.
    device - without /dev, e.g. "dm-10"
    mpath - without /dev/mapper, e.g. "mpatha"
    size - in bytes
    nodes - a list of nodes without /dev, e.g. ["sda","sdb"]
    """
    def createMP(self, device, mpath, size, nodes):
        for node in nodes:
            self._addDevice('/dev/'+node, size)
            self.dev_data['/dev/'+node]['mp_vol']= '/dev/mapper/'+mpath
        self._addDevice('/dev/'+device, size)
        self._addLink('/dev/'+device, '/dev/mapper/'+mpath)
        self._addVol(mpath, size, 1, 'mapper', ('/dev/'+n for n in nodes))
        self.vol_data['/dev/mapper/'+mpath]['type']='multipath'
        self.vol_data['/dev/mapper/'+mpath]['real_dev']='/dev/'+device
        self.dev_data['/dev/'+device]['major'] = misc.get_dmnumber('device-mapper')

    def setUp(self):
        super(MultipathFunctionCheck, self).setUp()
        self._options = main.Options()
        self._addPool('mapper', [])

        # some standard, non-mp devices
        self._addDevice('/dev/sdc', 2684354560)
        self._addDevice('/dev/sdc1', 2042177280, 1)
        self._addDevice('/dev/sdc2', 29826161, 2)
        self._addDevice('/dev/sdc3', 1042177280, 3)

        # two mp volumes
        self.createMP("dm-90", "mpatha", 11489037516, ["sda", "sdb"])
        self.createMP("dm-91", "mpathb", 11489037, ["sdd", "sde", "sdf"])


    def _mp_size(self, rawsize):
        size = rawsize / 1024


    def mock_run(self, cmd, *args, **kwargs):

        # Convert all parts of cmd into string
        for i, item in enumerate(cmd):
            if type(item) is not str:
                cmd[i] = str(item)

        self.run_data.append(" ".join(cmd))
        output = ""
        if cmd[:2] == ['multipath', '-ll']:
            mp_vol = None
            if len(cmd) > 2:
                mp_vol = cmd[2]
            counter=0
            for (vol, v_data) in self.vol_data.items():
                size_converted = misc.humanize_size(v_data['dev_size'])

                if size_converted[-1] == 'B':
                    size_converted = size_converted[:-1].replace(" ","")

                # if there was an optional argument (volume name),
                # let's continue if we are not in the correct volume
                mp_name = "mpath"+chr(ord('a')+counter)
                counter += 1
                if mp_vol != None and mp_vol != mp_name:
                    continue
                mp_id="XX360000000000000000e0000000"+chr(ord('a')+counter)
                output += "{name} ({id}) {dev} QEMU    ,QEMU HARDDISK \n".format(name=mp_name, id=mp_id, dev=basename(v_data['real_dev']))
                output += "size={0} features='0' hwhandler='0' wp=rw\n".format(size_converted)

                devs = []
                for (dev, d_data) in sorted(self.dev_data.items()):
                    if not 'mp_vol' in d_data or d_data['mp_vol'] != vol:
                        continue
                    devs.append(d_data)

                if len(devs):
                    for d_data in devs[:-1]:
                        output += "|-+- policy='service-time 0' prio=1 status=active\n"
                        # [5:] to drop the "/dev/" part from the path
                        output += "| `- 11:0:0:1 {0} 8:64 active ready running\n".format(d_data['dev_name'][5:])
                    # format the last item differently
                    d_data = devs[-1]
                    output += "`-+- policy='service-time 0' prio=1 status=enabled\n"
                    output += "  `- 11:0:0:1  {0} 8:64 active ready running\n".format(d_data['dev_name'][5:])
        if 'return_stdout' in kwargs and not kwargs['return_stdout']:
            output = None
        return (0, output, None)

    def test_mp_get_real_device(self):
        mp = MultipathDevice(options=self._options)
        self.assertEqual(mp.get_real_device("mpatha"), "/dev/dm-90")
        self.assertEqual(mp.get_real_device("mpathb"), "/dev/dm-91")
        self.assertEqual(mp.get_real_device("dm-90"), "/dev/dm-90")
        self.assertEqual(mp.get_real_device("dm-91"), "/dev/dm-91")
        self.assertEqual(mp.get_real_device("/dev/foo"), "/dev/foo")

    def test_mp_get_volumes_list(self):
        mp = MultipathDevice(options=self._options)
        self.assertEqual(mp.get_mp_devices(), ["mpatha", "mpathb"])

    def test_mp_get_device_data(self):
        mp = MultipathDevice(options=self._options)

        self.assertEqual(
            mp.get_device_data("/dev/sda", "mpatha", 0),
            {'dev_name':'/dev/sda', 'hide':False, 'multipath_volname':'mpatha',
             'pool_name':'/dev/dm-90', 'mount':'MULTIPATH'})
        self.assertEqual(
            mp.get_device_data("/dev/mapper/mpatha", "mpatha", 0),
            {'dev_name':'/dev/mapper/mpatha', 'hide':False,
             'multipath_volname':'mpatha', 'pool_name':'/dev/dm-90',
             'mount':'MULTIPATH'})

    def test_mp_get_volume_data(self):
        mp = MultipathDevice(options=self._options)
        vdata = mp.get_volume_data("mpatha")
        self.assertEqual(vdata['dev_name'], '/dev/dm-90')
        self.assertEqual(vdata['nodes'], ['/dev/sda','/dev/sdb'])

    def test_mp_forbidden_ops(self):
        self.assertRaises(problem.SsmError, main.main, "ssm remove /dev/mapper/mpatha")
        self.assertRaises(problem.SsmError, main.main, "ssm remove /dev/dm-90")
        self.assertRaises(problem.SsmError, main.main, "ssm remove mpatha")

    def test_mp_mock_raw_data(self):
        self._stdout = sys.stdout
        sys.stdout = self._stringio = StringIO()
        try:
            main.main("ssm list dev")
        finally:
            sys.stdout = self._stdout
        vol_entries = 0
        for line in self._stringio.getvalue().splitlines():
            if line[:6] in ['------', 'Device']:
                continue
            dev = line.split(" ")
            if dev[0] in ['/dev/dm-90', '/dev/dm-91']:
                vol_entries += 1
        self.assertEqual(vol_entries, 2, "List vol should show 2 entries for 2 multipath devices, but found {0}.".format(vol_entries))

        # There should be no output for pools
        sys.stdout = self._stringio = StringIO()
        try:
            main.main("ssm list pool")
        finally:
            sys.stdout = self._stdout
        self.assertEqual(self._stringio.getvalue(),
            "", "Multipath should show no pool, but has some.")

        # There should be no output for vols
        sys.stdout = self._stringio = StringIO()
        try:
            main.main("ssm list vol")
        finally:
            sys.stdout = self._stdout
        self.assertEqual(self._stringio.getvalue(),
            "", "Multipath should show no vols, but has some.")

        sys.stdout = self._stringio = StringIO()
        try:
            main.main("ssm list vol")
        finally:
            sys.stdout = self._stdout



#!/usr/bin/env python

import re
import sys, os

# Add into PATH so we don't use installed ssmlib, but the one we are developing
sys.path.insert(0, os.path.join( os.path.dirname(os.path.realpath(__file__)), ".." ))

from ssmlib import main

SYNOPSIS_INC = "src/synopsis.inc"
OPTIONS_DIR = "src/options/"
SSM_OPTIONS_INC = OPTIONS_DIR + "ssm_options.inc"
CREATE_OPTIONS_INC = OPTIONS_DIR + "create_options.inc"
INFO_OPTIONS_INC = OPTIONS_DIR + "info_options.inc"
LIST_OPTIONS_INC = OPTIONS_DIR + "list_options.inc"
REMOVE_OPTIONS_INC = OPTIONS_DIR + "remove_options.inc"
RESIZE_OPTIONS_INC = OPTIONS_DIR + "resize_options.inc"
CHECK_OPTIONS_INC = OPTIONS_DIR + "check_options.inc"
SNAPSHOT_OPTIONS_INC = OPTIONS_DIR + "snapshot_options.inc"
ADD_OPTIONS_INC = OPTIONS_DIR + "add_options.inc"
MOUNT_OPTIONS_INC = OPTIONS_DIR + "mount_options.inc"
MIGRATE_OPTIONS_INC = OPTIONS_DIR + "migrate_options.inc"

SSM_USAGE_INC = OPTIONS_DIR + "ssm_usage.inc"
CREATE_USAGE_INC = OPTIONS_DIR + "create_usage.inc"
INFO_USAGE_INC = OPTIONS_DIR + "info_usage.inc"
LIST_USAGE_INC = OPTIONS_DIR + "list_usage.inc"
REMOVE_USAGE_INC = OPTIONS_DIR + "remove_usage.inc"
RESIZE_USAGE_INC = OPTIONS_DIR + "resize_usage.inc"
CHECK_USAGE_INC = OPTIONS_DIR + "check_usage.inc"
SNAPSHOT_USAGE_INC = OPTIONS_DIR + "snapshot_usage.inc"
ADD_USAGE_INC = OPTIONS_DIR + "add_usage.inc"
MOUNT_USAGE_INC = OPTIONS_DIR + "mount_usage.inc"
MIGRATE_USAGE_INC = OPTIONS_DIR + "migrate_usage.inc"


class GenerateIncludes(object):

    def __init__(self):
        self.storage = main.StorageHandle()
        self.ssm_parser = main.SsmParser(self.storage, 'ssm')
        self.parser = self.ssm_parser.parser

    def _parse_usage(self, usage):
        s = re.compile('usage: |\n')
        res = s.sub(r'', usage)
        s = re.compile('\s\s+')
        res = s.sub(r' ', res)
        s = re.compile('(--\w+|-.)')
        res = s.sub(r'**\1**', res)
        s = re.compile('([a-z]+)\s+')
        res = s.sub(r'**\1** ', res)
        return res

    def _write_message(self, message, filename):
        with open(filename, 'w') as f:
            f.write(message)

    def format_synopsis(self, parser):
        return "{0}\n\n".format(self._parse_usage(parser.format_usage()))

    def write_ssm_usage(self):
        message = self.format_synopsis(self.ssm_parser.parser)
        self._write_message(message, SSM_USAGE_INC)

    def write_create_usage(self):
        message = self.format_synopsis(self.ssm_parser.parser_create)
        self._write_message(message, CREATE_USAGE_INC)

    def write_info_usage(self):
        message = self.format_synopsis(self.ssm_parser.parser_info)
        self._write_message(message, INFO_USAGE_INC)

    def write_list_usage(self):
        message = self.format_synopsis(self.ssm_parser.parser_list)
        self._write_message(message, LIST_USAGE_INC)

    def write_remove_usage(self):
        message = self.format_synopsis(self.ssm_parser.parser_remove)
        self._write_message(message, REMOVE_USAGE_INC)

    def write_resize_usage(self):
        message = self.format_synopsis(self.ssm_parser.parser_resize)
        self._write_message(message, RESIZE_USAGE_INC)

    def write_check_usage(self):
        message = self.format_synopsis(self.ssm_parser.parser_check)
        self._write_message(message, CHECK_USAGE_INC)

    def write_snapshot_usage(self):
        message = self.format_synopsis(self.ssm_parser.parser_snapshot)
        self._write_message(message, SNAPSHOT_USAGE_INC)

    def write_add_usage(self):
        message = self.format_synopsis(self.ssm_parser.parser_add)
        self._write_message(message, ADD_USAGE_INC)

    def write_mount_usage(self):
        message = self.format_synopsis(self.ssm_parser.parser_mount)
        self._write_message(message, MOUNT_USAGE_INC)

    def write_migrate_usage(self):
        message = self.format_synopsis(self.ssm_parser.parser_migrate)
        self._write_message(message, MIGRATE_USAGE_INC)

    def write_usage(self):
        self.write_ssm_usage()
        self.write_create_usage()
        self.write_info_usage()
        self.write_list_usage()
        self.write_remove_usage()
        self.write_resize_usage()
        self.write_check_usage()
        self.write_snapshot_usage()
        self.write_add_usage()
        self.write_mount_usage()
        self.write_migrate_usage()

    def _format_options(self, parser):
        help = parser.format_help()
        out = False
        message = ""
        for line in help.split('\n'):
            if line == '':
                out = False
            if out:
                message += "{0}\n".format(line[2:])
            if line == "optional arguments:":
                out = True
        return message

    def write_option_includes(self):
        message = self._format_options(self.parser)
        self._write_message(message, SSM_OPTIONS_INC)

        message = self._format_options(self.ssm_parser.parser_create)
        self._write_message(message, CREATE_OPTIONS_INC)

        message = self._format_options(self.ssm_parser.parser_info)
        self._write_message(message, INFO_OPTIONS_INC)

        message = self._format_options(self.ssm_parser.parser_list)
        self._write_message(message, LIST_OPTIONS_INC)

        message = self._format_options(self.ssm_parser.parser_remove)
        self._write_message(message, REMOVE_OPTIONS_INC)

        message = self._format_options(self.ssm_parser.parser_resize)
        self._write_message(message, RESIZE_OPTIONS_INC)

        message = self._format_options(self.ssm_parser.parser_check)
        self._write_message(message, CHECK_OPTIONS_INC)

        message = self._format_options(self.ssm_parser.parser_snapshot)
        self._write_message(message, SNAPSHOT_OPTIONS_INC)

        message = self._format_options(self.ssm_parser.parser_add)
        self._write_message(message, ADD_OPTIONS_INC)

        message = self._format_options(self.ssm_parser.parser_mount)
        self._write_message(message, MOUNT_OPTIONS_INC)

        message = self._format_options(self.ssm_parser.parser_migrate)
        self._write_message(message, MIGRATE_OPTIONS_INC)

includes = GenerateIncludes()

includes.write_option_includes()
includes.write_usage()



# (C)2011 Red Hat, Inc., Lukas Czerner <lczerner@redhat.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

# Miscellaneous functions for use by System Storage Manager

from __future__ import print_function

import os
import re
import sys
import stat
import tempfile
import threading
import subprocess
from ssmlib import problem
from base64 import encode

if sys.version < '3':
    def __next__(iter):
        return iter.next()
else:
    def __next__(iter):
        return next(iter)

# List of temporary mount points which should be cleaned up
# before exiting
TMP_MOUNTED = []

# A debug flag, because we can't reach to main.py from here
VERBOSE_VV_FLAG = False
VERBOSE_VVV_FLAG = False

if sys.version < '3':
    def __str__(x):
        if x is not None:
            return str(x)
else:
    def __str__(x):
        if x is not None:
            return str(x, encoding='utf-8', errors='strict')


def get_unit_size(string):
    """
    Check the last character of the string for the unit and return the tuple
    of unit value, unit name, otherwise return zero. It check only the first
    character of the unit string.

    >>> get_unit_size("B")
    (1, 'B')
    >>> get_unit_size("k")
    (1024, 'k')
    >>> get_unit_size("KiB")
    (1024, 'KiB')
    >>> get_unit_size("KB")
    (1024, 'KB')
    >>> get_unit_size("10M")
    (1048576, 'M')
    >>> get_unit_size("-99.99g")
    (1073741824, 'g')
    >>> get_unit_size("-99.99GiB")
    (1073741824, 'GiB')
    >>> get_unit_size("+99.99G")
    (1073741824, 'G')
    >>> get_unit_size("+99.99gb")
    (1073741824, 'gb')
    >>> get_unit_size("99.99g")
    (1073741824, 'g')
    >>> get_unit_size("0.84T")
    (1099511627776, 'T')
    >>> get_unit_size("0.84TiB")
    (1099511627776, 'TiB')
    >>> get_unit_size("0p")
    (1125899906842624, 'p')
    >>> get_unit_size("99kit")
    (0, '')
    >>> get_unit_size("")
    (0, '')
    >>> get_unit_size("H")
    (0, '')
    """

    mult = 0
    units = {'B': 1, 'K': 2 ** 10, 'M': 2 ** 20, 'G': 2 ** 30, 'T': 2 ** 40,
             'P': 2 ** 50}
    unit = re.sub(r'^\+?-?\d+(\.\d*)?', '', string)
    if unit and unit[0].upper() in units:
        mult = units[unit[0].upper()]
    all_units = ['B', 'K', 'M', 'G', 'T', 'P',
                 'KB', 'MB', 'GB', 'TB', 'PB',
                 'KIB', 'MIB', 'GIB', 'TIB', 'PIB']
    if unit.upper() in all_units:
        return mult, unit
    return 0, ""


def is_number(string):
    """
    Check is the string is number and return True or False.

    >>> is_number("3.14")
    True
    >>> is_number("+3.14")
    True
    >>> is_number("-3.14")
    True
    >>> is_number("314")
    True
    >>> is_number("3a14")
    False
    """
    try:
        float(string)
        return True
    except ValueError:
        return False


def get_real_size(size):
    """
    Get the real number from the size argument. It converts the size with units
    into the size in kilobytes. Is no unit is specified it defaults to
    kilobytes.

    >>> get_real_size("3141")
    '3141'
    >>> get_real_size("3141B")
    '3.07'
    >>> get_real_size("3141K")
    '3141.00'
    >>> get_real_size("3141KB")
    '3141.00'
    >>> get_real_size("3141KiB")
    '3141.00'
    >>> get_real_size("3141k")
    '3141.00'
    >>> get_real_size("3141M")
    '3216384.00'
    >>> get_real_size("3141G")
    '3293577216.00'
    >>> get_real_size("3141T")
    '3372623069184.00'
    >>> get_real_size("3141P")
    '3453566022844416.00'
    >>> get_real_size("3.14")
    '3.14'
    >>> get_real_size("+3.14")
    '+3.14'
    >>> get_real_size("-3.14")
    '-3.14'
    >>> get_real_size("3.14k")
    '3.14'
    >>> get_real_size("+3.14K")
    '+3.14'
    >>> get_real_size("-3.14k")
    '-3.14'
    >>> get_real_size("3.14G")
    '3292528.64'
    >>> get_real_size("3.14GB")
    '3292528.64'
    >>> get_real_size("3.14GiB")
    '3292528.64'
    >>> get_real_size("+3.14g")
    '+3292528.64'
    >>> get_real_size("-3.14G")
    '-3292528.64'
    >>> get_real_size("G")
    Traceback (most recent call last):
    ...
    Exception: Not supported unit in the size 'G' argument.
    >>> get_real_size("3141H")
    Traceback (most recent call last):
    ...
    Exception: Not supported unit in the size '3141H' argument.
    """
    if is_number(size):
        return size
    else:
        # Always use kilobytes in ssm
        mult, unit = get_unit_size(size)
        mult /= float(1024)
        number = re.sub(unit + "$", '', size)
        if is_number(number):
            sign = '+' if size[0] == '+' else ''
            if mult:
                return '{0}{1:.2f}'.format(sign, float(number) * mult)
    raise Exception("Not supported unit in the " +
                    "size \'{0}\' argument.".format(size))

def get_perc_size_argument(string):
    """
    Get percentage size argument. We now accept size argument in percentages
    of - free pool/volume space (FREE)
       - used pool/volume space (USED)
       - total pool/original volume size
    The accepted format is INTEGER%STRING where STRING needs to be one of
    the following (FREE, USED, or empty)
    """
    p = re.compile(r'%')
    perc, word = p.split(string, 1)

    if is_number(perc) and word.upper() in ['FREE', 'USED', '']:
        return (perc, word.upper())
    else:
        raise Exception("Not supported unit in the " +
                        "size \'{0}\' argument.".format(string))

def get_slaves(devname):
    return ["/dev/{0}".format(fname) for fname in os.listdir("/sys/block/{0}/slaves".format(devname))]


def send_udev_event(device, event):
    major, minor = get_major_minor(device)
    with open('/sys/dev/block/{0}:{1}/uevent'.format(major, minor), "w") as f:
        f.write(event)


def udev_settle():
    run(['udevadm', 'settle'], stderr=False, can_fail=True)


def get_device_by_uuid(uuid):
    path = "/dev/disk/by-uuid/{0}".format(uuid)
    return os.path.abspath(os.path.join(os.path.dirname(path),
                                        os.readlink(path)))


def get_major_minor(device):
    real_dev = get_real_device(device)
    info = os.stat(real_dev)
    major = os.major(info.st_rdev)
    minor = os.minor(info.st_rdev)
    return major, minor


def get_file_size(path):
    """
    Get size of the file (even block device) by seeking to the end of the
    file and returning offset. The returning size is in kilobytes.
    """
    with open(path, 'r') as f:
        return os.lseek(f.fileno(), os.SEEK_SET, os.SEEK_END) // 1024


def check_binary(name):
    command = ['which', name]
    if run(command, can_fail=True)[0]:
        return False
    return True


def do_mount(device, directory, options=None):
    command = ['mount']
    if options:
        command.extend(['-o', options])
    command.extend([device, directory])
    run(command)


def do_umount(mpoint, all_targets=False):
    command = ['umount']
    if all_targets:
        command.append('--all-targets')
    try:
        run(command + [mpoint])
    except RuntimeError:
        command.append('-l')
        run(command + mpoint)


def temp_mount(device, options=None):
    tmp = tempfile.mkdtemp()
    do_mount(device, tmp, options)
    TMP_MOUNTED.append(tmp)
    return tmp


def temp_umount(mpoint=None):
    if not mpoint:
        mpoint = TMP_MOUNTED.pop()
    do_umount(mpoint)
    os.rmdir(mpoint)


def do_cleanup():
    while 1:
        try:
            temp_umount()
        except IndexError:
            break


def get_signature(device, types=None):
    command = ["blkid", "-o", "value", "-p", "-s", "TYPE"]
    if types is not None:
        command.extend(['-u', types])
    command.append(device)

    ret, output, err = run(command, can_fail=True, stderr=False)
    output = output.strip()

    if ret:
        return None
    return output


def get_fs_type(device):
    return get_signature(device, "filesystem")


def get_real_device(device):
    if os.path.islink(device):
        return os.path.abspath(os.path.join(os.path.dirname(device),
                                            os.readlink(device)))
    return device


def get_swaps():
    swap = []
    with open('/proc/swaps', 'r') as f:
        for line in f.readlines()[1:]:
            swap.append(line.split())
    return swap


def get_partitions():
    partitions = []
    new_line = []
    output = run(["lsblk", "-l", "-b", "-n", "-p", "-o", "MAJ:MIN,SIZE,KNAME,NAME,PKNAME"],
                 stdout=False)

    for line in output[1].splitlines():
        new_line = re.split(r'\s+|:', line.strip())
        # Not every line has the parent device name, but in either case,
        # if we got data, convert the size to kB
        if len(new_line) in [5, 6]:
            new_line[2] = int(new_line[2])//1024
            partitions.append(new_line)
        else:
            pass
    return partitions


def get_mountinfo(regex=".*"):
    mounts = {}
    reg = re.compile(regex)
    names = ['id', 'parent', 'major_minor', 'root', 'mp', 'options']
    with open('/proc/self/mountinfo', 'r') as f:
        for line in f:
            m = reg.search(line)
            if not m:
                continue
            array = line.split(None, 6)
            row = dict([(names[index], array[index])
                        for index in min(
                            list(range(len(array) - 1)),
                            list(range(len(names)))
                        )])
            array = line.rsplit(None, 3)
            row['fs'] = array[1]
            row['dev'] = array[2]
            row['sb_options'] = array[3]
            dev = get_real_device(row['dev'])
            if row['root'] != '/':
                dev = "{0}:{1}".format(dev, row['root'])
            mounts[dev] = row
    return mounts

def get_mounts_old(regex=".*"):
    mounts = {}
    reg = re.compile(regex)
    with open('/proc/mounts', 'r') as f:
        for line in f:
            m = reg.search(line)
            if m:
                l = line.split()[:2]
                dev = get_real_device(l[0])
                mounts[dev] = {'dev': l[0], 'mp': l[1]}
    return mounts


def get_mounts(regex=".*"):
    if os.path.exists("/proc/self/mountinfo"):
        return get_mountinfo(regex)
    return get_mounts_old(regex)


def get_dmnumber(name):
    reg = re.compile(" {0}$".format(name))
    dmnumber = None
    with open('/proc/devices', 'r') as f:
        for line in f:
            m = reg.search(line)
            if m:
                dmnumber = line.split()[0]
                break
    return dmnumber

def udev_checkpoint(devices):
    if not isinstance(devices, list):
        devices = [devices]
    for dev in devices:
        send_udev_event(dev, "change")
    udev_settle()

def wipefs(devices, signatures):
    if not isinstance(devices, list):
        devices = [devices]
    if not isinstance(signatures, list):
        signatures = [signatures]
    command = ['wipefs', '-a', '-t', ','.join(signatures)] + devices
    # Avoid race with udev
    udev_settle()
    run(command)


def humanize_size(arg):
    """
    Returns the number with power of two units "KiB, MiB, ...etc. Parameter arg
    should be string of non-zero length, or integer. IMPORTANT: The arg
    argument is expected to be in KiB.

    >>> humanize_size(314)
    '314.00 KB'
    >>> humanize_size("314")
    '314.00 KB'
    >>> humanize_size(314159)
    '306.80 MB'
    >>> humanize_size(314159265)
    '299.61 GB'
    >>> humanize_size(314159265358)
    '292.58 TB'
    >>> humanize_size(314159265358979)
    '285.73 PB'
    >>> humanize_size(314159265358979323)
    '279.03 EB'
    >>> humanize_size(314159265358979323846)
    '272.49 ZB'
    >>> humanize_size(314159265358979323846264)
    '266.10 YB'
    >>> humanize_size(314159265358979323846264338)
    '266103.25 YB'
    >>> humanize_size(-314159265)
    '-299.61 GB'
    >>> humanize_size("")
    ''
    >>> humanize_size("hello world")
    Traceback (most recent call last):
        ...
    ValueError: could not convert string to float: hello world
    """
    count = 0
    if isinstance(arg, str) and not arg:
        return ""
    size = float(arg)
    while abs(size) >= 1024 and count < 7:
        size /= 1024
        count += 1
    units = ["KB", "MB", "GB", "TB", "PB", "EB", "ZB", "YB"]
    try:
        unit = units[count]
    except IndexError:
        unit = "???"
    return ("{0:.2f} {1}").format(size, unit)


def run(cmd, show_cmd=False, stdout=False, stderr=True, can_fail=False,
        stdin_data=None, return_stdout=True):

    stdin = None
    if stdin_data is not None:
        stdin = subprocess.PIPE

    if stderr:
        stderr = subprocess.STDOUT
    else:
        stderr = subprocess.PIPE

    if stdout:
        stdout = None
    else:
        stdout = subprocess.PIPE

    # Convert all parts of cmd into string
    for i, item in enumerate(cmd):
        if not isinstance(item, str):
            cmd[i] = str(item)

    if VERBOSE_VV_FLAG:
        print('executing command: {}'.format(' '.join(cmd)))

    proc = subprocess.Popen(cmd, stdout=stdout,
                            stderr=stderr, stdin=stdin, close_fds=True)

    output, error = proc.communicate(input=stdin_data)

    err_msg = "ERROR exit code {0} for running command: \"{1}\"".format(
              proc.returncode, " ".join(cmd))

    if proc.returncode != 0 and show_cmd:
        if output is not None:
            print(output)
        if error is not None:
            print(error)
        sys.stderr.write(err_msg + '\n')

    if proc.returncode != 0 and not can_fail:
        if output is not None:
            print(output)
        if error is not None:
            print(error)
        raise problem.CommandFailed(err_msg, exitcode=proc.returncode)

    if VERBOSE_VVV_FLAG:
        msg = "Exit: {}".format(proc.returncode)
        if error:
            msg += ", Error: {}".format(error)
        if output:
            msg += "\nOutput: {}".format(output)

        print(msg)

    if not return_stdout:
        output = None

    return (proc.returncode, __str__(output), __str__(error))


def chain(*iterables):
    """
    Make an iterator that returns elements from the first iterable until
    it is exhausted, then proceeds to the next iterable, until all of the
    iterables are exhausted. Used for treating consecutive sequences as a
    single sequence. This code has been taken from itertools python module.

    chain('ABC', 'DEF') --> A B C D E F
    """
    for it in iterables:
        for element in it:
            yield element


if sys.version < '3':
    def izip(*iterables):
        """
        Make an iterator that aggregates elements from each of the iterables.
        Like zip() except that it returns an iterator instead of a list. Used
        for lock-step iteration over several iterables at a time. This code has
        been taken from itertools python module (Python 2).

        izip('ABCD', 'xy') --> Ax By
        """
        iterators = map(iter, iterables)
        while iterators:
            yield tuple(map(next, iterators))
else:
    def izip(*iterables):
        """
        Make an iterator that aggregates elements from each of the iterables.
        Like zip() except that it returns an iterator instead of a list. Used
        for lock-step iteration over several iterables at a time. This code has
        been taken from itertools python module (Python 3).

        izip('ABCD', 'xy') --> Ax By
        """
        sentinel = object()
        iterators = [iter(it) for it in iterables]
        while iterators:
            result = []
            for it in iterators:
                elem = next(it, sentinel)
                if elem is sentinel:
                    return
                result.append(elem)
            yield tuple(result)


def compress(data, selectors):
    """
    Make an iterator that filters elements from data returning only those
    that have a corresponding element in selectors that evaluates to True.
    Stops when either the data or selectors iterables has been exhausted.
    This code has been taken from itertools python module.

    compress('ABCDEF', [1,0,1,0,1,1]) --> A C E F
    """
    return (d for d, s in izip(data, selectors) if s)


def permutations(iterable, r=None):
    """
    Return successive r length permutations of elements in the iterable.
    This code has been taken from itertools python module.

    permutations('ABCD', 2) --> AB AC AD BA BC BD CA CB CD DA DB DC
    permutations(range(3)) --> 012 021 102 120 201 210
    """
    pool = tuple(iterable)
    n = len(pool)
    r = n if r is None else r
    if r > n:
        return
    indices = list(range(n))
    cycles = list(range(n, n - r, -1))
    yield tuple(pool[i] for i in indices[:r])
    while n:
        for i in reversed(range(r)):
            cycles[i] -= 1
            if cycles[i] == 0:
                indices[i:] = indices[i + 1:] + indices[i:i + 1]
                cycles[i] = n - i
            else:
                j = cycles[i]
                indices[i], indices[-j] = indices[-j], indices[i]
                yield tuple(pool[i] for i in indices[:r])
                break
        else:
            return


def terminal_size(default=(25, 80)):
    """
    Returns running terminal size. If size cannot be found out default size
    is returned.
    """
    def _ioctl_GWINSZ(fd):
        try:
            import fcntl
            import termios
            import struct
            cr = struct.unpack('hh',
                               fcntl.ioctl(fd, termios.TIOCGWINSZ, '1234'))
        except:
            return None
        return cr

    try:
        env = os.environ
        cr = (env['LINES'], env['COLUMNS'])
    except KeyError:
        cr = _ioctl_GWINSZ(0) or _ioctl_GWINSZ(1) or _ioctl_GWINSZ(2)
        if not cr:
            try:
                fd = os.open(os.ctermid(), os.O_RDONLY)
                cr = _ioctl_GWINSZ(fd)
                os.close(fd)
            except:
                pass
        if not cr:
            cr = default
    return int(cr[1]), int(cr[0])


def is_bdevice(path):
    """
    Check whether the path is block device. If it is return
    the path to the real block device, otherwise return False
    """
    path = get_real_device(path)
    try:
        mode = os.lstat(path).st_mode
    except OSError:
        return False
    if not stat.S_ISBLK(mode):
        return False
    return path


def get_device_size(device):
    info = os.stat(device)
    major, minor = divmod(info.st_rdev, 256)
    with open("/sys/dev/block/{0}:{1}/size".format(major, minor), 'r') as f:
        for line in f:
            size = int(line)//2
            return size

def ptable(data, table_header=None):
    """
    Print data in a table, optionally with a header.
    The data has to be a list of tuples of strings [('a', 'b', 'c'), ...]
    The header is a tuple: (('name', type), ... ), where the type is used to decide alignment.
    Int and float aligns to right, anything else to left.
    All the tuples has to have the same number of members.
    """
    if len(data) == 0:
        return

    header = []
    types = []
    fmt = ""
    skip_header = True
    # Keep track of used columns. Then we only print out columns with values.
    columns = [False] * len(data[0])

    len_matrix = []
    if table_header:
        skip_header = False
        line = []
        for n, t in table_header:
            if not isinstance(n, str) or not isinstance(t, type):
                raise ValueError("The header for ptable has to be a tuple/list in the format: " +
                                 "[('name', type), ...], but got [..., ({}, {}), ...]".format(n, t))
            line.append(len(n))
            header.append(n)
            types.append(t)
        # add header lengths into the matrix
        len_matrix.append(line)
    else:
        header = [''] * len(data[0])
        types = [str] * len(data[0])

    for _ in data:
        len_matrix.append([0 for _ in data[0]])

    index = 0
    # Gather all lines which are going to be printed into the list
    # and create matrix of attribute lengths.
    # Iterate through all items.
    for itemsline in data:
        # a line
        for i, item in enumerate(itemsline):
            len_matrix[index][i] = len(item)
            if len(item) > 0:
                columns[i] = True
        index += 1


    if header:
        alignment = [(len(item)) for item in header]
    else:
        alignment = [0]*len(data[0])
        types = [str]*len(data[0])
    term_width = terminal_size()[0]

    # Update matrix of attribute lengths and construct the final list
    # of alignment for each column in the table.
    for index in range(len(len_matrix)):
        line = None
        # Find maximum length for each column
        for a, array in enumerate(len_matrix):
            for i, item in enumerate(array):
                if not columns[i]:
                    alignment[i] = 0
                    continue
                if item > alignment[i]:
                    alignment[i] = item
                    line = a

        # Check the overall line length and if it is longer then the
        # actual terminal width we can wrap the line right after the
        # first attribute. Simply set the alignment to the smaller
        # possible and let recalculate the list of column alignments.
        # Note that when even with the line wrap we would still exceed
        # the terminal width, then there is nothing we can do about it
        # so do not bother with line wrapping at all since it would
        # only screw the formatting even more.
        length = sum(alignment) + 2 * len(header) - 2
        if length > term_width and \
                (length - term_width) < (alignment[0] - len(header[0])) and \
                line is not None:
            alignment[0] = len(header[0])
            len_matrix[line][0] = len(header[0])
        else:
            break

    # Get the actual line width
    width = sum(compress(alignment, columns)) + 2 * len(header) - 2

    pos = 0
    # Use column alignments list to construct formatting string for each
    # line in the table. Note that some lines might be wrapped later on.
    for i, t in enumerate(types):
        if not columns[i]:
            continue
        if t in (float, int):
            fmt += "{{{0}:>{1}}}  ".format(pos, alignment[i])
        else:
            # Do not append additional spaces if this is the last item
            if i == len(header) - 1:
                fmt += "{{{0}:{1}}}".format(pos, alignment[i])
            else:
                fmt += "{{{0}:{1}}}  ".format(pos, alignment[i])
        pos += 1


    if not skip_header:
        print("-" * width)
        print(fmt.format(*tuple(header)))
    print("-" * width)
    # Now print each line of the table. When the first attribute of the
    # line is longer than it should be we know that we have to wrap the
    # line.
    for i, line in enumerate(data):
        line = compress(line, columns)
        tmp1 = __next__(line)
        if len(tmp1) > alignment[0]:
            print(tmp1)
            print(fmt.format('', *line))
        else:
            print(fmt.format(tmp1, *line))
    print("-" * width)


class Node(object):
    """ A simple graph node class """

    def __init__(self):
        self._neighbours = []
        self._parents = []
        self._children = []

    @property
    def neighbours(self):
        return self._neighbours

    @property
    def parents(self):
        return self._parents

    @property
    def children(self):
        return self._children

    def add_neighbour(self, node):
        """ A two-way method. It will add self into the other node as well. """
        if not isinstance(node, Node):
            raise ValueError("Only other Nodes can be added as a neighbour.")
        if not node in self._neighbours:
            self._neighbours.append(node)
            node.add_neighbour(self)

    def add_children(self, node):
        """ A two-way method. It will add self into the other node as well. """
        if not node in self._children:
            self.add_neighbour(node)
            self._children.append(node)
            node.add_parent(self)

    def add_parent(self, node):
        """ A two-way method. It will add self into the other node as well. """
        if not node in self._parents:
            self.add_neighbour(node)
            self._parents.append(node)
            node.add_children(self)

    def get_roots(self):
        """ Find all root nodes by recursively traversing up all the parents. """
        roots = set()
        for parent in self.parents:
            roots |= parent.get_roots()

        if not roots:
            return {self}
        return roots

    @staticmethod
    def find_node(name, node_cls):
        """ Search item of class node_cls and try to find one
            with specific name.
        """
        if node_cls is Pool:
            for item in pools:
                if item['pool_name'] == name:
                    return item

        elif node_cls is Volumes:
            for item in volumes:
                if item['dev_name'] == name:
                    return item

        elif node_cls is Devices:
            for item in devices:
                if item['dev_name'] == name:
                    return item

        elif node_cls is Snapshots:
            for item in snapshots:
                if item['dev_name'] == name:
                    return item
        return None



# (C)2013 Red Hat, Inc., Lukas Czerner <lczerner@redhat.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

# template module for System Storage Manager contains template classes
# to use when creating new backend

import os
from ssmlib import misc
from ssmlib import problem

__all__ = ["Backend", "BackendPool", "BackendVolume", "BackendDevice"]

try:
    SSM_TEMPLATE_DEFAULT_POOL = os.environ['SSM_TEMPLATE_DEFAULT_POOL']
except KeyError:
    SSM_TEMPLATE_DEFAULT_POOL = "template_pool"


class Backend(object):
    def __init__(self, options, data=None):
        self.type = 'template'
        self.data = data or {}
        self.options = options
        self.output = None
        self.default_pool_name = SSM_TEMPLATE_DEFAULT_POOL
        self.problem = problem.ProblemSet(options)

    def __str__(self):
        return repr(self.data)

    def __iter__(self):
        for item in sorted(self.data):
            yield item

    def __getitem__(self, key):
        if key in self.data:
            return self.data[key]


class BackendPool(Backend):
    def __init__(self, *args, **kwargs):
        super(BackendPool, self).__init__(*args, **kwargs)

    def reduce(self, pool, device):
        self.problem.check(self.problem.NOT_SUPPORTED,
                        ["Reducing pool", "{} backend".format(self.type)])

    def new(self, pool, devices):
        self.problem.check(self.problem.NOT_SUPPORTED,
                        ["Creating new pool", "{} backend".format(self.type)])

    def extend(self, pool, devices):
        self.problem.check(self.problem.NOT_SUPPORTED,
                        ["Extending pool", "{} backend".format(self.type)])

    def remove(self, pool):
        self.problem.check(self.problem.NOT_SUPPORTED,
                        ["Removing pool", "{} backend".format(self.type)])

    def create(self, pool, size=None, name=None, devs=None,
               options=None):
        self.problem.check(self.problem.NOT_IMPLEMENTED,
                        ["Creating volume", "{} backend".format(self.type)])

    def migrate(self, pool, sources, targets=None):
        self.problem.check(self.problem.NOT_IMPLEMENTED,
                        ["Migrating devices", "{} backend".format(self.type)])


class BackendVolume(Backend):
    def __init__(self, *args, **kwargs):
        super(BackendVolume, self).__init__(*args, **kwargs)

    def remove(self, volume):
        self.problem.check(self.problem.NOT_SUPPORTED,
                        ["Removing volume", "{} backend".format(self.type)])

    def resize(self, volume, size, resize_fs=True):
        self.problem.check(self.problem.NOT_SUPPORTED,
                        ["Resizing volume", "{} backend".format(self.type)])


class BackendDevice(Backend):
    def __init__(self, *args, **kwargs):
        super(BackendDevice, self).__init__(*args, **kwargs)

    def remove(self, devices):
        self.problem.check(self.problem.NOT_SUPPORTED,
                        ["Removing device", "{} backend".format(self.type)])



# (C)2011 Red Hat, Inc., Lukas Czerner <lczerner@redhat.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

# lvm module for System Storage Manager

import re
import os
import stat
import datetime
from ssmlib import misc
from ssmlib import problem
from ssmlib.backends import template

__all__ = ["PvsInfo", "VgsInfo", "LvsInfo", "ThinPool"]

try:
    SSM_LVM_DEFAULT_POOL = os.environ['SSM_LVM_DEFAULT_POOL']
except KeyError:
    SSM_LVM_DEFAULT_POOL = "lvm_pool"

try:
    DM_DEV_DIR = os.environ['DM_DEV_DIR']
except KeyError:
    DM_DEV_DIR = "/dev"
MAX_LVS = 999

# TODO: This is ugly and needs to be removed and done properly
THIN_POOL_DATA = {}

def get_lvm_version():
    try:
        output = misc.run(['lvm', 'version'], can_fail=True)[1]
        output = output.strip().split("\n")
        pattern = re.compile("LVM version:")
        version = [0, 0, 0]
        for line in output:
            if pattern.match(line.strip()):
                match = " ".join(line.split())
                tmp = re.search(r'(?<=LVM version: )\d+\.\d+\.\d+',
                                    match).group(0)
                version = list(map(int, tmp.split(".", 3)))
    except (OSError, AttributeError):
        version = [0, 0, 0]
    return version

LVM_VERSION = get_lvm_version()

def create_thin_volume(parent_pool, thin_pool, virtsize, lvname):
    pool_volume = parent_pool + '/' + thin_pool

    # Ignore options for non existing thin volumes somehow
    command = ['lvcreate', '-n', lvname, '-T', pool_volume,
               '-V', str(virtsize) + 'K']
    command.insert(0, "lvm")
    misc.run(command, stdout=True)
    return "{0}/{1}/{2}".format(DM_DEV_DIR, parent_pool, lvname)


class LvmInfo(template.Backend):

    def __init__(self, *args, **kwargs):
        super(LvmInfo, self).__init__(*args, **kwargs)
        self.type = 'lvm'
        self.attrs = []
        self.binary = misc.check_binary('lvm')
        self.default_pool_name = SSM_LVM_DEFAULT_POOL
        self.init_local_problem_set()

    def parse_attr(self, lv, attr):
        if attr[4] == 'a':
            lv['active'] = True
        else:
            lv['active'] = False

    def init_local_problem_set(self):
        self.DEVICE_INACTIVE = \
            ['Device \'{0}\' is not active! It might contain filesystem we\'re unable to detect and it may be destroyed by this operation.',
             problem.PROMPT_CONTINUE,
             problem.FL_DEFAULT_NO | problem.FL_EXIT_ON_NO | problem.FL_FORCE_YES,
             problem.GeneralError]

    def run_lvm(self, command, noforce=False):
        if not self.binary:
            self.problem.check(self.problem.TOOL_MISSING, 'lvm')
        if self.options.force and not noforce:
            command.insert(1, "-f")
        if self.options.verbose:
            command.insert(1, "-v")
        command.insert(0, "lvm")
        misc.run(command, stdout=True)

    def _data_index(self, row):
        return row.values()[len(row.values()) - 1]

    def _skip_data(self, row):
        return False

    def _parse_data(self, command):
        if not self.binary:
            return
        ret, self.output, err = misc.run(command, stderr=False, can_fail=True)
        # A workaround for LVM behaviour:
        # lvm lvs' exit code is 5 on exported volumes, even if everything
        # is ok. So, if the code is 5, command was 'lvm lvs ...'
        # and error message says that a volume was exported, ignore the
        # error
        if ret != 0:
            err_msg = "ERROR exit code {0} for running command: \"{1}\"".format(
                      ret, " ".join(command))
            if ret != 5 or command[0:2] != ['lvm', 'lvs'] or \
               not str(err).endswith('is exported\n'):
                if err is not None:
                    print(err)
                raise problem.CommandFailed(err_msg, exitcode=ret)

        for line in self.output.split("\n"):
            if not line:
                break
            array = line.split("|")
            row = dict([(self.attrs[index], array[index].lstrip())
                       for index in range(len(array))])
            if self._skip_data(row):
                continue
            self._fill_aditional_info(row)
            self.data[self._data_index(row)] = row

    def _fill_aditional_info(self, row):
        pass

    def supported_since(self, version, string):
        if version > LVM_VERSION:
            msg = "ERROR: You need at least lvm version " + \
                  "{0}. Feature \"{1}\"".format(".".join(map(str, version)),
                                                string)
            self.problem.check(self.problem.NOT_SUPPORTED, msg)
        return True

    def require_thin_support(self):

        # Old versions of lvm had a bug for some cases of thin provisioning.
        # Because Debian 8 still has an affected version in its repository,
        # check the version and prevent any thin provisioning related
        # operations.
        self.supported_since([2,2,112],"thin provisioning")

        # Test if thin provisioning tools are installed.
        # SSM is not directly using them, but lvm does. ome distributions
        # have it only as an optional dependency for lvm2 and if it is not installed,
        # lvm behaves strangely and can fail without any useful information
        # in middle of a sequence of commands SSM does.
        found = False
        for path in os.environ["PATH"].split(os.pathsep):
            path = path.strip('"')
            if os.path.isfile(os.path.join(path, 'thin_check')):
                    found = True
                    break

        if not found:
            msg = "ERROR: lvm does not have installed thin provisioning tools. " +\
                  "Some distributions mark it as an optional dependency for lvm2, " +\
                  "in which case, you need to install it manually"
            self.problem.check(self.problem.GENERAL_ERROR, msg)

        return True


class VgsInfo(LvmInfo, template.BackendPool):

    def __init__(self, *args, **kwargs):
        super(VgsInfo, self).__init__(*args, **kwargs)
        command = ["lvm", "vgs", "--separator", "|", "--noheadings",
                   "--nosuffix", "--units", "k", "-o",
                   "vg_name,pv_count,vg_size,vg_free,lv_count"]
        self.attrs = ['pool_name', 'dev_count', 'pool_size', 'pool_free',
                      'vol_count']

        self._parse_data(command)

    def _fill_aditional_info(self, vg):
        vg['type'] = 'lvm'
        vg['pool_used'] = float(vg['pool_size']) - float(vg['pool_free'])

    def _data_index(self, row):
        return row['pool_name']

    def _generate_lvname(self, lvname, vg):
        for i in range(1, MAX_LVS):
            name = "{0}{1:0>{align}}".format(lvname, i, align=len(str(MAX_LVS)))
            path = "{0}/{1}/{2}".format(DM_DEV_DIR, vg, name)
            if "{}/{}".format(vg, name) in THIN_POOL_DATA:
                continue
            try:
                if stat.S_ISBLK(os.stat(path).st_mode):
                    continue
            except OSError:
                pass
            return name
        self.problem.error("Can not find proper lvname!")

    def reduce(self, vg, device):
        command = ['vgreduce', vg, device]
        self.run_lvm(command)

    def new(self, vg, devices):
        if type(devices) is not list:
            devices = [devices]
        command = ['vgcreate', vg]
        command.extend(devices)
        self.run_lvm(command)

    def extend(self, vg, devices):
        if type(devices) is not list:
            devices = [devices]
        command = ['vgextend', vg]
        command.extend(devices)
        self.run_lvm(command)

    def remove(self, vg):
        command = ['vgremove', vg]
        self.run_lvm(command)

    def create(self, vg, size=None, name=None, devs=None,
               options=None):
        options = options or {}
        devices = devs or []
        command = ['lvcreate', vg]

        if 'virtsize' in options:
            # We're going to create new lv which is going to be thin pool
            # so name it appropriately
            lvname = self._generate_lvname(vg + "_thin", vg)
        elif name:
            lvname = name
        else:
            lvname = self._generate_lvname("lvol", vg)

        if 'virtsize' in options:
            self.require_thin_support()
            command.extend(['-T'])

        if size:
            command.extend(['-L', str(size) + 'K'])
        else:
            if len(devices) > 0:
                tmp = "100%PVS"
            else:
                tmp = "100%FREE"
            command.extend(['-l', tmp])

        command.extend(['-n', lvname.rpartition("/")[-1]])

        if 'raid' in options:
            if options['raid'] == '0':
                if not options['stripesize']:
                    options['stripesize'] = "64"
                if not options['stripes'] and len(devices) > 0:
                    options['stripes'] = str(len(devices))
                if not options['stripes']:
                    self.problem.error("Devices or number of " +
                                       "stripes should be defined!")
                if options['stripesize']:
                    command.extend(['-I', options['stripesize']])
                if options['stripes']:
                    command.extend(['-i', options['stripes']])
            elif options['raid'] == '1' and \
                 self.supported_since([2,2,89],"raid1"):
                if options['stripesize'] or options['stripes']:
                    msg = "ERROR: Specifying stripe size or number of " + \
                          "stripes when creating raid1 volume with lvm backend"
                    self.problem.check(self.problem.NOT_SUPPORTED, msg)
                # Unfortunately 50%PVS does not work here because it does not
                # take in account metadata needed to create mirrored volume.
                # Using 49%PVS is not viable either because it will cut off
                # a lot of potential storage. So we'll require to specify
                # size in this case.
                if not size:
                    msg = "ERROR: Creating raid1 with lvm backend without " + \
                          "specifying size"
                    self.problem.check(self.problem.NOT_SUPPORTED, msg)
                command.extend(["--type", "raid1"])
            elif options['raid'] == '10' and \
                 self.supported_since([2,2,98],"raid10"):
                if not options['stripesize']:
                    options['stripesize'] = "64"
                if not options['stripes'] and len(devices) > 0:
                    if len(devices) < 4:
                        self.problem.error("Number of devices should be at " +
                                           "least 4 in raid 10 setup")
                    if len(devices) % 2 != 0:
                        self.problem.error("Number of devices should be " +
                                           "even in raid 10 setup")
                    options['stripes'] = str(len(devices)//2)
                if not options['stripes']:
                    self.problem.error("Devices or number of " +
                                       "stripes should be defined")
                if int(options['stripes']) < 2:
                    self.problem.error("Number of stripes should be at " +
                                       "least 2 in raid 10 setup")
                if options['stripesize']:
                    command.extend(['-I', options['stripesize']])
                if options['stripes']:
                    command.extend(['-i', options['stripes']])
                if not size:
                    msg = "ERROR: Creating raid10 with lvm backend without " + \
                          "specifying size"
                    self.problem.check(self.problem.NOT_SUPPORTED, msg)
                command.extend(["--type", "raid10"])
            else:
                self.problem.not_supported("RAID level {0}".format(options['raid']) +
                                           " with \"lvm\" backend")

        # This should be done only if necessary - for example when the size
        # was not provided and is used as %PVS, otherwise let lvm decide
        # which devices to use from the pool.
        command.extend(devices)
        self.run_lvm(command, noforce=True)
        path = "{0}/{1}/{2}".format(DM_DEV_DIR, vg, lvname)
        if 'virtsize' in options:
            thin_pool = lvname
            if name:
                lvname = name
            else:
                lvname = self._generate_lvname("tvol", vg)
            path = create_thin_volume(vg, thin_pool, options['virtsize'], lvname)
        return path

    def migrate(self, vg, source, target):
        """ Migrate a PV to a new device using pvmove.

        Parameters
        ----------
        vg : [str]
            Volume group into which the PV belongs.
        source : [main.DeviceItem]
            Source PV device.
        target : [str]
            Path to the target device.
        """
        # We're ready to do pvmove source and target must be already in the
        # pool

        # pvmove does not accept -f, temporarily unset it
        force = self.options.force
        self.options.force = False

        # If the source is not used we do not have to do anything
        if float(source['dev_used']) > 0.0:
            command = ['pvmove', '--atomic', source.name, target ]
            self.run_lvm(command)

        misc.send_udev_event(source.name, "change")
        misc.send_udev_event(target, "change")

        # restore saved value
        self.options.force = force


class PvsInfo(LvmInfo, template.BackendDevice):

    def __init__(self, *args, **kwargs):
        super(PvsInfo, self).__init__(*args, **kwargs)
        command = ["lvm", "pvs", "--separator", "|", "--noheadings",
                   "--nosuffix", "--units", "k", "-o",
                   "pv_name,vg_name,pv_free,pv_used,pv_size"]
        self.attrs = ['dev_name', 'pool_name', 'dev_free',
                      'dev_used', 'dev_size']

        self._parse_data(command)

    def _data_index(self, row):
        return misc.get_real_device(row['dev_name'])

    def _fill_aditional_info(self, pv):
        pv['hide'] = False
        # If the device is not in any group we do not need this info
        # and we do not want it to show up in the device listing
        if not pv['pool_name']:
            pv['dev_used'] = ''
            pv['dev_free'] = ''
            del pv['pool_name']

    def remove(self, devices):
        if len(devices) == 0:
            return
        command = ['pvremove']
        command.extend(devices)
        self.run_lvm(command)


class LvsInfo(LvmInfo, template.BackendVolume):

    def __init__(self, *args, **kwargs):
        super(LvsInfo, self).__init__(*args, **kwargs)
        command = ["lvm", "lvs", "--separator", "|", "--noheadings",
                   "--nosuffix", "--units", "k", "-o",
                   "vg_name,lv_size,stripes,stripesize,segtype,lv_name,origin,lv_attr,pool_lv"]
        self.attrs = ['pool_name', 'vol_size', 'stripes',
                      'stripesize', 'type', 'lv_name', 'origin', 'attr', 'pool_lv']
        self.handle_fs = True
        self.mounts = misc.get_mounts('{0}/mapper'.format(DM_DEV_DIR))
        self._parse_data(command)

    def _fill_aditional_info(self, lv):
        lv['dev_name'] = "{0}/{1}/{2}".format(DM_DEV_DIR, lv['pool_name'],
                                              lv['lv_name'])
        if lv['origin'] or \
           lv['attr'][0] == 't':
            lv['hide'] = True

        # Show thin-pool as a pool name in case of thin volumes
        if lv['type'] == 'thin':
            lv['parent_pool'] = lv['pool_name']
            lv['pool_name'] = lv['pool_lv']

        lv['real_dev'] = misc.get_real_device(lv['dev_name'])

        sysfile = "/sys/block/{0}/dm/name".format(
                  os.path.basename(lv['real_dev']))

        # In some weird cases the "real" device might not be in /dev/dm-*
        # form (see tests). In this case constructed sysfile will not exist
        # so we just use real device name to search mounts.
        try:
            with open(sysfile, 'r') as f:
                lvname = f.readline()[:-1]
            lv['dm_name'] = "{0}/mapper/{1}".format(DM_DEV_DIR, lvname)
        except IOError:
            lv['dm_name'] = lv['real_dev']

        if lv['real_dev'] in self.mounts:
            lv['mount'] = self.mounts[lv['real_dev']]['mp']
        self.parse_attr(lv, lv['attr'])

    def __getitem__(self, name):
        if name in self.data:
            return self.data[name]
        if DM_DEV_DIR + "/" + name in self.data:
            return self.data[DM_DEV_DIR + "/" + name]
        device = name
        if not os.path.exists(name):
            device = DM_DEV_DIR + "/" + name
            if not os.path.exists(device):
                return None
        device = misc.get_real_device(device)
        if device in self.data:
            return self.data[device]
        return None

    def _data_index(self, row):
        return row['real_dev']

    def _get_dev_name(self, lv):
        real = misc.get_real_device(lv)
        if real in self.data:
            return self.data[real]['dev_name']
        else:
            return lv

    def remove(self, lv):
        vol = self[lv]
        if 'mount' in vol:
            if self.problem.check(self.problem.FS_MOUNTED,
                                  [vol['dev_name'], vol['mount']]):
                misc.do_umount(vol['mount'])
        lv = self._get_dev_name(lv)
        command = ['lvremove', lv]
        self.run_lvm(command)

    def resize(self, lv, size, resize_fs=True):
        lv = self._get_dev_name(lv)
        command = ['lvresize', '-L', str(size) + 'k', lv]
        vol = self[lv]
        if vol['active'] == False and \
           size < float(vol['vol_size']):
            self.problem.check(self.DEVICE_INACTIVE, lv)
        if resize_fs:
            command.insert(1, '-r')
        self.run_lvm(command)

    def snapshot(self, lv, destination, name, snap_size=None):
        vol = self[lv]
        vol_size = float(vol['vol_size'])
        lv = self._get_dev_name(lv)

        if not name:
            now = datetime.datetime.now()
            name = now.strftime("snap%Y%m%dT%H%M%S")

        if vol['type'] == "thin":
            if snap_size:
                self.problem.warn("Setting snapshot size for thin volume is" +
                                  " not supported")
                snap_size = None
        elif not snap_size:
            # We'll ceate snapshot of the size of 20% of the original volume
            snap_size = vol_size * 0.20

        if snap_size:
            command = ['lvcreate', '--size', str(snap_size) + 'K', '--snapshot',
                       '--name', name, lv]
        else:
            command = ['lvcreate', '--snapshot', '--name', name, lv]

        self.run_lvm(command)


class SnapInfo(LvmInfo):

    def __init__(self, *args, **kwargs):
        super(SnapInfo, self).__init__(*args, **kwargs)
        command = ["lvm", "lvs", "--separator", "|", "--noheadings",
                   "--nosuffix", "--units", "k", "-o",
                   "vg_name,lv_size,stripes,stripesize,segtype," +
                   "lv_name,origin,snap_percent,lv_attr,pool_lv"]
        self.attrs = ['pool_name', 'vol_size', 'stripes',
                      'stripesize', 'type', 'lv_name', 'origin',
                      'snap_size', 'attr', 'pool_lv']
        self.handle_fs = True
        self.mounts = misc.get_mounts('{0}/mapper'.format(DM_DEV_DIR))
        self._parse_data(command)

    def _skip_data(self, row):
        if not row['origin']:
            return True
        else:
            return False

    def _data_index(self, row):
        return misc.get_real_device(row['dev_name'])

    def _fill_aditional_info(self, snap):
        snap['dev_name'] = "{0}/{1}/{2}".format(DM_DEV_DIR, snap['pool_name'],
                                                snap['lv_name'])
        snap['hide'] = False
        snap['snap_path'] = snap['dev_name']

        if snap['type'] != "thin":
            # It's possible that snap_percent in lvm output is not defined.
            # For example on inactive volume, so just remove it.
            if snap['snap_size']:
                size = float(snap['vol_size']) * float(snap['snap_size'])
                snap['snap_size'] = str(size / 100.00)
            else:
                del snap['snap_size']
        else:
            # Show thin-pool as a pool name in case of thin volumes
            snap['parent_pool'] = snap['pool_name']
            snap['pool_name'] = snap['pool_lv']

        snap['real_dev'] = misc.get_real_device(snap['dev_name'])

        sysfile = "/sys/block/{0}/dm/name".format(
                  os.path.basename(snap['real_dev']))

        # In some weird cases the "real" device might not be in /dev/dm-*
        # form (see tests). In this case constructed sysfile will not exist
        # so we just use real device name to search mounts.
        try:
            with open(sysfile, 'r') as f:
                lvname = f.readline()[:-1]
            snap['dm_name'] = "{0}/mapper/{1}".format(DM_DEV_DIR, lvname)
        except IOError:
            snap['dm_name'] = snap['real_dev']

        if snap['real_dev'] in self.mounts:
            snap['mount'] = self.mounts[snap['real_dev']]['mp']

        self.parse_attr(snap, snap['attr'])


class ThinPool(LvmInfo, template.BackendPool):

    def __init__(self, *args, **kwargs):
        super(ThinPool, self).__init__(*args, **kwargs)
        self.type = 'thin'
        command = ["lvm", "lvs", "--separator", "|", "--noheadings",
                   "--nosuffix", "--units", "k", "-o",
                   "vg_name,lv_size,stripes,stripesize,segtype,lv_name," +
                   "origin,lv_attr,pv_count,thin_count,data_percent," +
                   "metadata_percent,snap_percent"]
        self.attrs = ['parent_pool', 'vol_size', 'stripes',
                      'stripesize', 'type', 'lv_name', 'origin', 'attr',
                      'dev_count', 'vol_count', 'data_percent',
                      'metadata_percent', 'snap_percent']
        self._parse_data(command)
        # Uff, so ugly...needs to be changed
        global THIN_POOL_DATA
        THIN_POOL_DATA = self.data

    def _fill_aditional_info(self, vg):
        vg['type'] = 'thin'
        vg['pool_name'] = os.path.basename(vg['lv_name'])
        vg['index_name'] = "{}/{}".format(vg['parent_pool'], vg['pool_name'])
        vg['pool_size'] = vg['vol_size']
        vg['pool_used'] = float(vg['vol_size']) * (float(vg['data_percent'])/100)
        vg['pool_free'] = float(vg['vol_size']) - vg['pool_used']
        if vg['attr'][4] == 'a':
            vg['active'] = True
        else:
            vg['active'] = False

    def __getitem__(self, key):
        # we can have <key> in self.data and then it is simple
        # but we can also have <parentpool>/<key> and then it gets complicated
        if key in self.data:
            return self.data[key]

        found = None
        for (name,item) in self.data.items():
            subkey = name.split('/', 1)
            if subkey[1] == key:
                if not found:
                    found = item
                else:
                    raise Exception("Multiple items with name {} found".format(key))
        return found

    def _data_index(self, row):
        return row['index_name']

    def _skip_data(self, row):
        if row['attr'][0] not in ['t', 'T']:
            return True
        else:
            return False

    def _generate_lvname(self, lvname, vg):
        for i in range(1, MAX_LVS):
            name = "{0}{1:0>{align}}".format(lvname, i, align=len(str(MAX_LVS)))
            path = "{0}/{1}/{2}".format(DM_DEV_DIR, vg, name)
            if name in self.data:
                continue
            try:
                if stat.S_ISBLK(os.stat(path).st_mode):
                    continue
            except OSError:
                pass
            return name
        self.problem.error("Can not find proper lvname!")

    def reduce(self, vg, device):
        msg = "Removing devices from thin pool"
        self.problem.check(self.problem.NOT_SUPPORTED, msg)

    def new(self, vg, devices):
        msg = "Creating a new"
        self.problem.check(self.problem.NOT_SUPPORTED, msg)

    def extend(self, vg, devices):
        self.require_thin_support()
        # Add devices to the parent pool first
        vg = self[vg]
        pool = vg['parent_pool']
        if type(devices) is not list:
            devices = [devices]
        command = ['vgextend', pool]
        command.extend(devices)
        self.run_lvm(command)
        # Now resize the thin-pool volume
        lv = vg['parent_pool'] + '/' + vg['lv_name']
        command = ['lvresize', lv]
        command.extend(devices)
        if vg['active'] == False:
            self.problem.check(self.DEVICE_INACTIVE, lv)
        self.run_lvm(command)

    def remove(self, vg):
        vg = self[vg]
        lvname = vg['parent_pool'] + '/' + vg['lv_name']
        command = ['lvremove', lvname]
        self.run_lvm(command)

    def create(self, vg, size=None, name=None, devs=None,
               options=None):
        self.require_thin_support()
        vg = self[vg]
        if vg['active'] == False:
            lv = vg['parent_pool'] + '/' + vg['lv_name']
            self.problem.check(self.DEVICE_INACTIVE, lv)

        if name:
            lvname = name
        else:
            lvname = self._generate_lvname("tvol", vg['parent_pool'])

        pool_volume = vg['parent_pool'] + '/' + vg['lv_name']

        # Once can specify either --size or --virtual-size argument when
        # creating thin volume out of thin pool, but not both.
        if size and 'virtsize' in options:
            self.problem.error("Either '--size' or '--virtual-size' can be" +
                               " can be specified for new thin volume size" +
                               " but not both")

        virtsize = None
        # Size needs to be specified for the thin volume
        if 'virtsize' in options:
            virtsize = options['virtsize']
        elif size:
            virtsize = size
        else:
            self.problem.error("Size must be specified to create a volume " +
                               "from a thin pool")

        # Ignore options for non existing thin volumes somehow
        command = ['lvcreate', '-n', lvname, '-T', pool_volume,
                   '-V', str(virtsize) + 'K']
        self.run_lvm(command)
        return "{0}/{1}/{2}".format(DM_DEV_DIR, vg['parent_pool'], lvname)



#!/usr/bin/env python
#
# (C)2011 Red Hat, Inc., Lukas Czerner <lczerner@redhat.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

# Main testing script for the system storage manager


import os
import re
import sys
import time
import doctest
import unittest
import argparse
import distutils.spawn

os.environ['SSM_NONINTERACTIVE'] = "1"

from ssmlib import main
from ssmlib import misc
from ssmlib.backends import lvm, crypt, btrfs, multipath

import tests.unittests as tests_module
from tests.unittests import *


def prog_exists(program):
    """ Test if given program/path/file exists and can be executed.
        Will search in $PATH too.

    Parameters
    ----------
    program : str
        Absolute/relative path to search for.

    Returns
    -------
    bool
        True if the program exists and can be executed.
    """

    return bool(distutils.spawn.find_executable(program))


def check_system_dependencies():
    """ Verify if we have all dependencies in system installed and print any
        that is missing.
        This function will not test for python modules available, only binaries.

    Returns
    -------
    bool
        True if everything is ok, False if any dependency is missing.
    """

    binaries = [
        'blkid',
        'which',
        'mount',
        'wipefs',
        'dd',
        'tune2fs',
        'resize2fs',
        'mkfs.ext3',
        'mkfs.ext4',
        'mkfs.xfs',
        'mkfs.btrfs',
        'fsck.ext3',
        'fsck.ext4',
        'fsck.xfs',
        'fsck.btrfs',
        'xfs_db',
        'xfs_growfs',
        'xfs_repair',
        'lvm',
        'dmsetup',
        'cryptsetup',
        'multipath'
    ]

    missing = []
    for prog in binaries:
        if not prog_exists(prog):
            missing.append(prog)

    if missing:
        print("Dependencies are missing; some tests will probably fail or won't run:")
        print('\n'.join(missing))
        return False
    return True


def run_bash_tests(names, want_logs=False):
    cur = os.getcwd()
    os.chdir('./tests/bashtests')
    command = ['ls', '-m']
    if os.access('.coverage', os.R_OK):
        os.remove('.coverage')

    failed = []
    passed = []
    count = 0
    misc.run('./set.sh', stdout=False)
    output = misc.run(command, stdout=False)[1]
    t0 = time.time()
    for script in output.split(","):
        script = script.strip()
        if not re.match("^\d\d\d-.*\.sh$", script):
            continue
        if names and script not in names and script[:3] not in names:
            continue
        count += 1
        sys.stdout.write("{0:<29}".format(script) + " ")
        sys.stdout.flush()
        bad_file = re.sub("\.sh$",".bad", script)
        if os.access(bad_file, os.R_OK):
            os.remove(bad_file)
        ret, out, err = misc.run(['./' + script], stdout=False, can_fail=True)
        if ret:
            print("\033[91m[FAILED]\033[0m")
            failed.append(script)
            with open(bad_file, 'w') as f:
                f.write(out)
        elif re.search("Traceback", out):
            # There should be no tracebacks in the output
            out += "\nWARNING: Traceback in the output!\n"
            print("\033[93m[WARNING]\033[0m")
            with open(bad_file, 'w') as f:
                f.write(out)
        else:
            print("\033[92m[PASSED]\033[0m")
            passed.append(script)

    if count == 0 and names:
        print("[+] No bash test matches the name(s)")
        return 0

    t1 = time.time() - t0

    if want_logs:
        print_logs(failed)

    print("Ran {0} tests in {1} seconds.".format(count, round(t1, 2)))
    print("{0} tests PASSED: {1}".format(len(passed), ", ".join(passed)))
    ret = 0
    if len(failed) > 0:
        print("{0} tests FAILED: {1}".format(len(failed), ", ".join(failed)))
        print("See files with \"bad\" extension for output")
        ret = 1
    # Show coverage report output if possible
    if misc.check_binary('coverage'):
        print("[+] Coverage")
        misc.run(['coverage', 'report'], stdout=True, can_fail=True)
    os.chdir(cur)
    return ret

def print_logs(tests):
    for test in tests:
        logfile = re.sub(r"\.sh$",".bad", test)
        print("-----------------------------------\n{}\n-----------------------------------".format(
            logfile
        ))
        try:
            with open(logfile, 'r') as f:
                print(f.read())
        except IOError:
            print("The file '{}' does not exist.".format(logfile))
    print("-----------------------------------")


def doc_tests():
    print("[+] Running doctests")
    doctest_flags = doctest.IGNORE_EXCEPTION_DETAIL | doctest.ELLIPSIS | \
                    doctest.REPORT_ONLY_FIRST_FAILURE
    result = doctest.testmod(main, exclude_empty=True, report=True,
            raise_on_error=False, optionflags=doctest_flags)
    result = doctest.testmod(lvm, exclude_empty=True, report=True,
            raise_on_error=False, optionflags=doctest_flags)
    result = doctest.testmod(crypt, exclude_empty=True, report=True,
            raise_on_error=False, optionflags=doctest_flags)
    result = doctest.testmod(btrfs, exclude_empty=True, report=True,
            raise_on_error=False, optionflags=doctest_flags)
    result = doctest.testmod(multipath, exclude_empty=True, report=True,
            raise_on_error=False, optionflags=doctest_flags)
    result = doctest.testmod(misc, exclude_empty=True, report=True,
            raise_on_error=False, optionflags=doctest_flags)

def unit_tests(names):
    print("[+] Running unittests")
    tests = unittest.TestSuite()
    test_loader = unittest.TestLoader()

    if names:
        for name in names:
            if name[-3:] == ".sh" or name.isdigit():
                # bash test, skip here
                continue

            # first try a full name
            try:
                tests = unittest.TestSuite([tests, test_loader.loadTestsFromName(
                    name)])
                continue
            except (ImportError, AttributeError):
                pass

            # then a name with the prefix omitted
            try:
                tests = unittest.TestSuite([tests, test_loader.loadTestsFromName(
                    "tests.unittests." + name)])
                continue
            except (ImportError, AttributeError):
                pass

            # ok, maybe even the file name was omitted, so we are down to class.method name
            tests_classes = [cname for cname in dir(tests_module) if cname[:5] == 'test_']
            found = False
            for c in tests_classes:
                try:
                    tests = unittest.TestSuite([tests, test_loader.loadTestsFromName(
                        "tests.unittests.{}.{}".format(c,name))])
                    found = True
                    break
                except (ImportError, AttributeError):
                    pass
            if found:
                continue

            # still nothing found... it might be a method only, but TODO that
            print("Warning: Test {} was not found.".format(name))

        if tests.countTestCases() == 0:
            print("[+] No unittest matches the name(s)")
            return
    else:
        tests_lvm = test_loader.loadTestsFromModule(test_lvm)
        tests_btrfs = test_loader.loadTestsFromModule(test_btrfs)
        tests_ssm = test_loader.loadTestsFromModule(test_ssm)
        tests_misc = test_loader.loadTestsFromModule(test_misc)
        tests_multipath = test_loader.loadTestsFromModule(test_multipath)
        tests = unittest.TestSuite([tests_lvm, tests_btrfs, tests_ssm, tests_misc, tests_multipath])

    test_runner = unittest.TextTestRunner(verbosity=2)
    return not test_runner.run(tests).wasSuccessful()


if __name__ == '__main__':
    result = 0
    parser = argparse.ArgumentParser(description="Run the test suite for SSM. "
            "If both --bash and --unit arguments are ommited, run both groups. "
            "If a test name is specified, only matching tests are run.")
    parser.add_argument('-b', '--bash', dest='bash', action='store_true',
                    help='run only bash tests')
    parser.add_argument('-u', '--unit', dest='unit', action='store_true',
                    help='run only unit tests')
    parser.add_argument('-l', '--logs', dest='want_logs', action='store_true',
                    help='if a bash test fails, print out it\'s log to stdout')
    parser.add_argument('tests', metavar='TEST', type=str, nargs='*',
                    help='Specific tests to be run. For bash tests, '
                         'that means either a full name (001-foo.sh), '
                         'or just the number. '
                         'For unit tests, it means something like '
                         'BtrfsFunctionCheck.test_btrfs_resize for a specific test, '
                         'BtrfsFunctionCheck for specific test suite '
                         'and test_btrfs for a whole file of tests.')

    args = parser.parse_args()

    check_system_dependencies()

    run_all = not args.unit and not args.bash
    if args.unit and args.bash:
        print("Do not use both --bash and --unit at once."
            "All tests are run when these options are omitted.")
        sys.exit(1)

    if args.unit or run_all:
        if not args.tests:
            doc_tests()
        result = unit_tests(args.tests)
        if result:
            # if a unittest failed, break out immediately and do not try bash tests
            sys.exit(result)


    if args.bash or run_all:
        if not os.geteuid() == 0:
            print("\nRoot privileges required to run more tests!\n")
            sys.exit(0)
        print("[+] Running bash tests")
        result = run_bash_tests(names=args.tests, want_logs=args.want_logs)
    sys.exit(result)



# (C)2011 Red Hat, Inc., Lukas Czerner <lczerner@redhat.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

# crypt module for System Storage Manager

import re
import os
import stat
import tempfile
from ssmlib import misc
from ssmlib import problem
from ssmlib.backends import template

__all__ = ["DmCryptVolume"]

SUPPORTED_CRYPT = ['luks', 'plain']
CRYPT_SIGNATURES = ['crypto_LUKS']
CRYPT_DEFAULT_EXTENSION = "luks"

try:
    SSM_CRYPT_DEFAULT_POOL = os.environ['SSM_CRYPT_DEFAULT_POOL']
except KeyError:
    SSM_CRYPT_DEFAULT_POOL = "crypt_pool"

try:
    SSM_CRYPT_DEFAULT_VOL_PREFIX = os.environ['SSM_CRYPT_DEFAULT_VOL_PREFIX']
except KeyError:
    SSM_CRYPT_DEFAULT_VOL_PREFIX = "encrypted"

# cryptsetup against my expectations does not take into account
# DM_DEV_DIR so set it to /dev pernamently for now.
#try:
#    DM_DEV_DIR = os.environ['DM_DEV_DIR']
#except KeyError:
#    DM_DEV_DIR = "/dev"
DM_DEV_DIR = "/dev"
MAX_DEVS = 999


def get_cryptsetup_version():
    try:
        output = misc.run(['cryptsetup', '--version'], can_fail=True)[1]
        version = list(map(int, output.strip().split()[-1].split('.', 3)))
    except (OSError, AttributeError):
        version = [0, 0, 0]
    return version

CRYPTSETUP_VERSION = get_cryptsetup_version()


class DmObject(template.Backend):
    def __init__(self, *args, **kwargs):
        super(DmObject, self).__init__(*args, **kwargs)
        self.type = 'crypt'
        self.mounts = misc.get_mounts('{0}/mapper'.format(DM_DEV_DIR))
        self.default_pool_name = SSM_CRYPT_DEFAULT_POOL

        if not misc.check_binary('dmsetup') or \
           not misc.check_binary('cryptsetup'):
            return

    def run_cryptsetup(self, command, stdout=True, password=None):
        if not misc.check_binary('cryptsetup'):
            self.problem.check(self.problem.TOOL_MISSING, 'cryptsetup')
        command.insert(0, "cryptsetup")
        if password != None:
            return misc.run(command, stdout=stdout, stdin_data=password)
        else:
            return misc.run(command, stdout=stdout)


class DmCryptPool(DmObject, template.BackendPool):
    def __init__(self, *args, **kwargs):
        super(DmCryptPool, self).__init__(*args, **kwargs)
        '''
        pool = {'pool_name': self.default_pool_name,
                'type': 'crypt',
                'dev_count': '0',
                'pool_free': '0',
                'pool_used': '0',
                'pool_size': '0',
                'hide': True}
        self.data[self.default_pool_name] = pool
        '''
        pool = {}
        pool['type'] = 'crypt'
        pool['pool_name'] = self.default_pool_name
        pool['hide'] = True
        pool['pool_free'] = 0
        pool['pool_size'] = 0
        pool['pool_used'] = 0
        self.data[pool['pool_name']] = pool
        self.passphrase = None
        self.force_weak_password = False

    def set_passphrase(self, passphrase, force=False):
        self.passphrase = passphrase.encode()
        self.force_weak_password = force

    def create(self, pool, size=None, name=None, devs=None,
               options=None):
        if CRYPTSETUP_VERSION < [1, 6, 0]:
            msg = "You need at least cryptsetup version " + \
                  "{0}. Creating encrypted volumes".format('1.6.0')
            self.problem.check(self.problem.NOT_SUPPORTED, msg)
        options = options or {}
        if 'encrypt' in options:
            if options['encrypt'] is True:
                options['encrypt'] = CRYPT_DEFAULT_EXTENSION
            if options['encrypt'] not in SUPPORTED_CRYPT:
                self.problem.not_supported("Extension "
                                           "\'{0}\'".format(options['encrypt']))
        else:
            # So the options is not crypt specific. It's ok, just use defaults
            options['encrypt'] = CRYPT_DEFAULT_EXTENSION

        if len(devs) > 1:
            self.problem.not_supported("Device concatenation" +
                                       " with \"crypt\" backend")
        if not name:
            name = self._generate_devname()
        device = devs[0]
        args = []
        command = []
        if self.options.verbose:
            args.append('-v')
        else:
            args.append('-q')
        if options['encrypt'] == "luks":
            command.extend(args)
            if self.options.force or self.force_weak_password:
                command.append('--force-password')
            if self.options.interactive and not self.passphrase:
                command.append('-y')
            command.extend(['luksFormat', device])
            self.run_cryptsetup(command, password=self.passphrase)
        command = []
        command.extend(args)
        command.append('open')
        if size:
            # Size is in KiB but cryptsetup accepts it in 512 byte blocks
            size = str(float(size) * 2).split('.')[0]
            command.extend(['--size', size])
        command.extend(['--type', options['encrypt'], device, name])
        self.run_cryptsetup(command, password=self.passphrase)
        return "{0}/mapper/{1}".format(DM_DEV_DIR, name)

    def _generate_devname(self):
        for i in range(1, MAX_DEVS):
            name = "{0}{1:0>{align}}".format(SSM_CRYPT_DEFAULT_VOL_PREFIX, i,
                                            align=len(str(MAX_DEVS)))
            path = "{0}/mapper/{1}".format(DM_DEV_DIR, name)
            try:
                if stat.S_ISBLK(os.stat(path).st_mode):
                    continue
            except OSError:
                pass
            return name
        self.problem.error("Can not find proper device name. Specify one!")


class DmCryptVolume(DmObject, template.BackendVolume):

    def __init__(self, *args, **kwargs):
        super(DmCryptVolume, self).__init__(*args, **kwargs)

        command = ['dmsetup', 'table']
        self.output = misc.run(command, stderr=False)[1]
        for line in self.output.split("\n"):
            if not line or line == "No devices found":
                break
            dm = {}
            array = line.split()
            if len(array) == 1:
                continue
            dm['type'] = array[3]
            if dm['type'] != 'crypt':
                continue
            dm['vol_size'] = str(int(array[2]) / 2.0)
            devname = re.sub(":$", "",
                             "{0}/mapper/{1}".format(DM_DEV_DIR, array[0]))
            dm['dm_name'] = devname
            dm['pool_name'] = self.default_pool_name
            dm['dev_name'] = devname
            dm['real_dev'] = misc.get_real_device(devname)
            if dm['real_dev'] in self.mounts:
                dm['mount'] = self.mounts[dm['real_dev']]['mp']

            # Check if the device really exists in the system. In some cases
            # (tests) DM_DEV_DIR can lie to us, if that is the case, simple
            # ignore the device.
            if not os.path.exists(devname):
                continue
            command = ['cryptsetup', 'status', devname]
            self._parse_cryptsetup(command, dm)
            self.data[dm['dev_name']] = dm

    def _parse_cryptsetup(self, cmd, dm):
        self.output = misc.run(cmd, stderr=False)[1]
        for line in self.output.split("\n"):
            if not line:
                break
            array = line.split()
            if array[0].strip() == 'cipher:':
                dm['cipher'] = array[1]
            elif array[0].strip() == 'keysize:':
                dm['keysize'] = array[1]
            elif array[0].strip() == 'device:':
                dm['crypt_device'] = array[1]

    def __getitem__(self, name):
        if name in self.data:
            return self.data[name]
        device = name
        if not os.path.exists(name):
            device = DM_DEV_DIR + "/" + name
            if not os.path.exists(device):
                return None
        device = misc.get_real_device(device)
        if device in self.data:
            return self.data[device]
        return None

    def remove(self, dm):
        vol = self[dm]
        if 'mount' in vol:
            if self.problem.check(self.problem.FS_MOUNTED,
                                  [vol['dev_name'], vol['mount']]):
                misc.do_umount(vol['mount'])
        command = ['remove', dm]
        self.run_cryptsetup(command)
        misc.wipefs(vol['crypt_device'], CRYPT_SIGNATURES)


class DmCryptDevice(DmObject, template.BackendDevice):

    def __init__(self, *args, **kwargs):
        super(DmCryptDevice, self).__init__(*args, **kwargs)

        for line in misc.get_partitions():
            device = {}
            devname = line[3]
            signature = misc.get_signature(devname)
            if signature in CRYPT_SIGNATURES:
                device['hide'] = False
                device['dev_name'] = devname
                device['pool_name'] = self.default_pool_name
                device['dev_free'] = '0'
                device['dev_used'] = str(misc.get_device_size(devname))
                self.data[devname] = device

    def remove(self, devices):
        misc.wipefs(devices, CRYPT_SIGNATURES)
